import unittest

from numpy import array

import Orange.classification.naive_bayes as nb
from Orange import preprocess
from Orange.data.sql.table import SqlTable
from Orange.data import Domain
from Orange.data.variable import DiscreteVariable
from Orange.tests.sql.base import DataBaseTest as dbt


class NaiveBayesTest(unittest.TestCase, dbt):
    def setUpDB(self):
        self.conn, self.iris = self.create_iris_sql_table()

    def tearDownDB(self):
        self.drop_iris_sql_table()

    @dbt.run_on(["postgres"])
    def test_NaiveBayes(self):
        iris_v = ['Iris-setosa', 'Iris-virginica', 'Iris-versicolor']
        table = SqlTable(self.conn, self.iris,
                         type_hints=Domain([],
                                           DiscreteVariable("iris",
                                                            values=iris_v)))
        disc = preprocess.Discretize()
        table = disc(table)
        bayes = nb.NaiveBayesLearner()
        clf = bayes(table)
        # Single instance prediction
        self.assertEqual(clf(table[0]), table[0].get_class())
        # Table prediction
        pred = clf(table)
        actual = array([ins.get_class() for ins in table])
        ca = pred == actual
        ca = ca.sum() / len(ca)
        self.assertGreater(ca, 0.95)
        self.assertLess(ca, 1.)



import warnings
from functools import partial
from itertools import chain

import numpy as np

from AnyQt.QtWidgets import QHeaderView, QStyledItemDelegate, QMenu
from AnyQt.QtGui import QStandardItemModel, QStandardItem
from AnyQt.QtCore import Qt, QSize, QObject, pyqtSignal as Signal, \
    QSortFilterProxyModel
from sklearn.exceptions import UndefinedMetricWarning

from Orange.data import Variable, DiscreteVariable, ContinuousVariable
from Orange.evaluation import scoring
from Orange.widgets import gui
from Orange.widgets.gui import OWComponent
from Orange.widgets.settings import Setting


def check_results_adequacy(results, error_group, check_nan=True):
    error_group.add_message("invalid_results")
    error_group.invalid_results.clear()

    def anynan(a):
        return np.any(np.isnan(a))

    if results is None:
        return None
    if results.data is None:
        error_group.invalid_results(
            "Results do not include information on test data")
    elif not results.data.domain.has_discrete_class:
        error_group.invalid_results(
            "Discrete outcome variable is required")
    elif not results.actual.size:
        error_group.invalid_results(
            "Empty result on input. Nothing to display.")
    elif check_nan and (anynan(results.actual) or
                        anynan(results.predicted) or
                        (results.probabilities is not None and
                         anynan(results.probabilities))):
        error_group.invalid_results(
            "Results contains invalid values")
    else:
        return results


def results_for_preview(data_name=""):
    from Orange.data import Table
    from Orange.evaluation import CrossValidation
    from Orange.classification import \
        LogisticRegressionLearner, SVMLearner, NuSVMLearner

    data = Table(data_name or "heart_disease")
    results = CrossValidation(
        data,
        [LogisticRegressionLearner(penalty="l2"),
         LogisticRegressionLearner(penalty="l1"),
         SVMLearner(probability=True),
         NuSVMLearner(probability=True)
        ],
        store_data=True
    )
    results.learner_names = ["LR l2", "LR l1", "SVM", "Nu SVM"]
    return results


BUILTIN_SCORERS_ORDER = {
    DiscreteVariable: ("AUC", "CA", "F1", "Precision", "Recall"),
    ContinuousVariable: ("MSE", "RMSE", "MAE", "R2")}


def learner_name(learner):
    """Return the value of `learner.name` if it exists, or the learner's type
    name otherwise"""
    return getattr(learner, "name", type(learner).__name__)


def usable_scorers(target: Variable):
    order = {name: i
             for i, name in enumerate(BUILTIN_SCORERS_ORDER[type(target)])}
    # 'abstract' is retrieved from __dict__ to avoid inheriting
    usable = (cls for cls in scoring.Score.registry.values()
              if cls.is_scalar and not cls.__dict__.get("abstract")
              and isinstance(target, cls.class_types))
    return sorted(usable, key=lambda cls: order.get(cls.name, 99))


def scorer_caller(scorer, ovr_results, target=None):
    def thunked():
        with warnings.catch_warnings():
            # F-score and Precision return 0 for labels with no predicted
            # samples. We're OK with that.
            warnings.filterwarnings(
                "ignore", "((F-score|Precision)) is ill-defined.*",
                UndefinedMetricWarning)
            if scorer.is_binary:
                return scorer(ovr_results, target=target, average='weighted')
            else:
                return scorer(ovr_results)

    return thunked


class ScoreModel(QSortFilterProxyModel):
    def lessThan(self, left, right):
        def is_bad(x):
            return not isinstance(x, (int, float, str)) \
                   or isinstance(x, float) and np.isnan(x)

        left = left.data()
        right = right.data()
        is_ascending = self.sortOrder() == Qt.AscendingOrder

        # bad entries go below; if both are bad, left remains above
        if is_bad(left) or is_bad(right):
            return is_bad(right) == is_ascending

        # for data of different types, numbers are at the top
        if type(left) is not type(right):
            return isinstance(left, float) == is_ascending

        # case insensitive comparison for strings
        if isinstance(left, str):
            return left.upper() < right.upper()

        # otherwise, compare numbers
        return left < right


class ScoreTable(OWComponent, QObject):
    shown_scores = \
        Setting(set(chain(*BUILTIN_SCORERS_ORDER.values())))

    shownScoresChanged = Signal()

    class ItemDelegate(QStyledItemDelegate):
        def sizeHint(self, *args):
            size = super().sizeHint(*args)
            return QSize(size.width(), size.height() + 6)

        def displayText(self, value, locale):
            if isinstance(value, float):
                return f"{value:.3f}"
            else:
                return super().displayText(value, locale)

    def __init__(self, master):
        QObject.__init__(self)
        OWComponent.__init__(self, master)

        self.view = gui.TableView(
            wordWrap=True, editTriggers=gui.TableView.NoEditTriggers
        )
        header = self.view.horizontalHeader()
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        header.setDefaultAlignment(Qt.AlignCenter)
        header.setStretchLastSection(False)
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(self.show_column_chooser)

        self.model = QStandardItemModel(master)
        self.model.setHorizontalHeaderLabels(["Method"])
        self.sorted_model = ScoreModel()
        self.sorted_model.setSourceModel(self.model)
        self.view.setModel(self.sorted_model)
        self.view.setItemDelegate(self.ItemDelegate())

    def _column_names(self):
        return (self.model.horizontalHeaderItem(section).data(Qt.DisplayRole)
                for section in range(1, self.model.columnCount()))

    def show_column_chooser(self, pos):
        # pylint doesn't know that self.shown_scores is a set, not a Setting
        # pylint: disable=unsupported-membership-test
        def update(col_name, checked):
            if checked:
                self.shown_scores.add(col_name)
            else:
                self.shown_scores.remove(col_name)
            self._update_shown_columns()

        menu = QMenu()
        header = self.view.horizontalHeader()
        for col_name in self._column_names():
            action = menu.addAction(col_name)
            action.setCheckable(True)
            action.setChecked(col_name in self.shown_scores)
            action.triggered.connect(partial(update, col_name))
        menu.exec(header.mapToGlobal(pos))

    def _update_shown_columns(self):
        # pylint doesn't know that self.shown_scores is a set, not a Setting
        # pylint: disable=unsupported-membership-test
        header = self.view.horizontalHeader()
        for section, col_name in enumerate(self._column_names(), start=1):
            header.setSectionHidden(section, col_name not in self.shown_scores)
        self.view.resizeColumnsToContents()
        self.shownScoresChanged.emit()

    def update_header(self, scorers):
        # Set the correct horizontal header labels on the results_model.
        self.model.setColumnCount(3 + len(scorers))
        self.model.setHorizontalHeaderItem(0, QStandardItem("Model"))
        self.model.setHorizontalHeaderItem(1, QStandardItem("Train time [s]"))
        self.model.setHorizontalHeaderItem(2, QStandardItem("Test time [s]"))
        for col, score in enumerate(scorers, start=3):
            item = QStandardItem(score.name)
            item.setToolTip(score.long_name)
            self.model.setHorizontalHeaderItem(col, item)
        self._update_shown_columns()



"""
A modified pyqtgraph/exporters/SVGExporter.py from
https://github.com/pyqtgraph/pyqtgraph/ (available under MIT License)

Fixes:
- SVGExporter: fix axes position and scale
  (https://github.com/pyqtgraph/pyqtgraph/pull/641)
- SVG export: handle Qt.NoPen on Qt5
  (https://github.com/pyqtgraph/pyqtgraph/pull/642)

Remove this file when pyqtgraph updates.
"""

from orangewidget.utils.SVGExporter import SVGExporter

__all__ = ["SVGExporter"]



import numpy as np

from Orange.data import Domain, Table, ContinuousVariable
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.unsupervised.owkmeans import OWKMeans

from .base import benchmark


def table(rows, cols):
    return Table.from_numpy(  # pylint: disable=W0201
        Domain([ContinuousVariable(str(i)) for i in range(cols)]),
        np.random.RandomState(0).rand(rows, cols))


class BenchOWKmeans(WidgetTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.d_100_100 = table(100, 100)
        cls.d_sampled_silhouette = table(10000, 1)
        cls.d_10_500 = table(10, 500)

    def setUp(self):
        self.widget = None  # to avoid lint errors

    def widget_from_to(self):
        self.widget = self.create_widget(
            OWKMeans, stored_settings={"auto_commit": False})
        self.widget.controls.k_from.setValue(2)
        self.widget.controls.k_to.setValue(6)

    @benchmark(number=3, warmup=1, repeat=3)
    def bench_from_to_100_100(self):
        self.widget_from_to()
        self.send_signal(self.widget.Inputs.data, self.d_100_100)
        self.commit_and_wait(wait=100*1000)

    @benchmark(number=3, warmup=1, repeat=3)
    def bench_from_to_100_100_no_normalize(self):
        self.widget_from_to()
        self.widget.normalize = False
        self.send_signal(self.widget.Inputs.data, self.d_100_100)
        self.commit_and_wait(wait=100*1000)

    @benchmark(number=3, warmup=1, repeat=3)
    def bench_from_to_sampled_silhouette(self):
        self.widget_from_to()
        self.send_signal(self.widget.Inputs.data, self.d_sampled_silhouette)
        self.commit_and_wait(wait=100*1000)

    @benchmark(number=3, warmup=1, repeat=3)
    def bench_wide(self):
        self.widget = self.create_widget(
            OWKMeans, stored_settings={"auto_commit": False})
        self.send_signal(self.widget.Inputs.data, self.d_10_500)
        self.commit_and_wait(wait=100*1000)



import unittest


class TestAxisItem(unittest.TestCase):
    def test_scalable_axis_item(self):
        from pyqtgraph import __version__
        # When upgraded to 0.11.1 (or bigger) check if resizing AxisItem font
        # works and overwritten functions generateDrawSpecs and
        # _updateMaxTextSize are no longer needed.
        self.assertLess(__version__, "0.11.1")


if __name__ == '__main__':
    unittest.main()



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest
from unittest.mock import Mock

import numpy as np
import scipy.sparse as sp
from scipy.sparse import csr_matrix, csc_matrix

from Orange.data import DiscreteVariable, Table, Domain
from Orange.statistics import contingency
from Orange import data
from Orange.tests import test_filename


def assert_dist_equal(dist, expected):
    np.testing.assert_array_equal(np.asarray(dist), expected)


def assert_dist_almost_equal(dist, expected):
    np.testing.assert_array_almost_equal(np.asarray(dist), expected)


class TestDiscrete(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.zoo = data.Table("zoo")
        cls.test9 = data.Table(test_filename("datasets/test9.tab"))

    def test_discrete(self):
        cont = contingency.Discrete(self.zoo, 0)
        assert_dist_equal(cont["amphibian"], [4, 0])
        assert_dist_equal(cont, [[4, 0], [20, 0], [13, 0],
                                 [4, 4], [10, 0], [2, 39], [5, 0]])

        cont = contingency.Discrete(self.zoo, "predator")
        assert_dist_equal(cont["fish"], [4, 9])
        assert_dist_equal(cont, [[1, 3], [11, 9], [4, 9], [7, 1],
                                 [2, 8], [19, 22], [1, 4]])

        cont = contingency.Discrete(self.zoo, self.zoo.domain["predator"])
        assert_dist_equal(cont["fish"], [4, 9])
        assert_dist_equal(cont, [[1, 3], [11, 9], [4, 9], [7, 1],
                                 [2, 8], [19, 22], [1, 4]])
        self.assertEqual(sum(cont.col_unknowns), 0)
        self.assertEqual(sum(cont.row_unknowns), 0)

    def test_discrete_missing(self):
        d = data.Table("zoo")
        d.Y[25] = float("nan")
        d[0][0] = float("nan")
        cont = contingency.Discrete(d, 0)
        assert_dist_equal(cont["amphibian"], [3, 0])
        assert_dist_equal(cont, [[3, 0], [20, 0], [13, 0], [4, 4],
                                 [10, 0], [2, 38], [5, 0]])
        np.testing.assert_almost_equal(cont.col_unknowns,
                                       [0, 0, 0, 0, 0, 1, 0])
        np.testing.assert_almost_equal(cont.row_unknowns,
                                       [1, 0])

        d = data.Table("zoo")
        d.Y[2] = float("nan")
        d[2]["predator"] = float("nan")
        cont = contingency.Discrete(d, "predator")
        assert_dist_equal(cont["fish"], [4, 8])
        assert_dist_equal(cont, [[1, 3], [11, 9], [4, 8], [7, 1],
                                 [2, 8], [19, 22], [1, 4]])
        np.testing.assert_almost_equal(
            cont.col_unknowns, [0, 0, 0, 0, 0, 0, 0])
        np.testing.assert_almost_equal(cont.row_unknowns, [0, 0])
        self.assertEqual(1, cont.unknowns)

    def test_array_with_unknowns(self):
        d = data.Table("zoo")
        d.Y[2] = float("nan")
        d.Y[6] = float("nan")
        d[2]["predator"] = float("nan")
        d[4]["predator"] = float("nan")
        cont = contingency.Discrete(d, "predator")
        assert_dist_equal(cont.array_with_unknowns,
                          [[1, 3, 0], [11, 9, 0], [4, 8, 0], [7, 1, 0],
                           [2, 8, 0], [18, 21, 1], [1, 4, 0], [1, 0, 1]])

    def test_discrete_with_fallback(self):
        d = data.Table("zoo")
        d.Y[25] = None
        d.Y[24] = None
        d.X[0, 0] = None
        d.X[24, 0] = None
        default = contingency.Discrete(d, 0)

        d._compute_contingency = Mock(side_effect=NotImplementedError)
        fallback = contingency.Discrete(d, 0)

        np.testing.assert_array_equal(
            np.asarray(fallback), np.asarray(default))
        np.testing.assert_array_equal(fallback.unknowns, default.unknowns)
        np.testing.assert_array_equal(
            fallback.row_unknowns, default.row_unknowns)
        np.testing.assert_array_equal(
            fallback.col_unknowns, default.col_unknowns)

    def test_continuous(self):
        d = data.Table("iris")
        cont = contingency.Continuous(d, "sepal width")
        correct = [[2.3, 2.9, 3.0, 3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7,
                    3.8, 3.9, 4.0, 4.1, 4.2, 4.4],
                   [1, 1, 6, 5, 5, 2, 9, 6, 2, 3, 4, 2, 1, 1, 1, 1]]

        np.testing.assert_almost_equal(cont.col_unknowns, [0, 0, 0])
        np.testing.assert_almost_equal(cont.row_unknowns, np.zeros(23))
        np.testing.assert_almost_equal(cont["Iris-setosa"], correct)
        self.assertEqual(cont.unknowns, 0)

        correct = [[2.2, 2.5, 2.6, 2.7, 2.8, 2.9, 3.0, 3.1, 3.2, 3.3, 3.4, 3.6, 3.8],
                   [1, 4, 2, 4, 8, 2, 12, 4, 5, 3, 2, 1, 2]]
        np.testing.assert_almost_equal(
            cont[d.domain.class_var.values.index("Iris-virginica")], correct)
        np.testing.assert_almost_equal(cont.col_unknowns, [0, 0, 0])
        np.testing.assert_almost_equal(cont.row_unknowns, np.zeros(23))
        self.assertEqual(cont.unknowns, 0)

    def test_continuous_missing(self):
        d = data.Table("iris")
        d[1][1] = float("nan")
        cont = contingency.Continuous(d, "sepal width")
        correct = [[2.3, 2.9, 3.0, 3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7,
                    3.8, 3.9, 4.0, 4.1, 4.2, 4.4],
                   [1, 1, 5, 5, 5, 2, 9, 6, 2, 3, 4, 2, 1, 1, 1, 1]]
        np.testing.assert_almost_equal(cont.col_unknowns, [1, 0, 0])
        np.testing.assert_almost_equal(cont.row_unknowns, np.zeros(23))
        np.testing.assert_almost_equal(cont["Iris-setosa"], correct)
        self.assertEqual(cont.unknowns, 0)

        d.Y[0] = float("nan")
        cont = contingency.Continuous(d, "sepal width")
        correct = [[2.2, 2.5, 2.6, 2.7, 2.8, 2.9, 3.0, 3.1, 3.2, 3.3, 3.4, 3.6, 3.8],
                   [1, 4, 2, 4, 8, 2, 12, 4, 5, 3, 2, 1, 2]]
        np.testing.assert_almost_equal(
            cont[d.domain.class_var.values.index("Iris-virginica")], correct)
        np.testing.assert_almost_equal(cont.col_unknowns, [1, 0, 0])
        np.testing.assert_almost_equal(
            cont.row_unknowns,
            [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.,
             0., 0., 0., 0., 0., 0., 0.])
        self.assertEqual(cont.unknowns, 0)

        d.Y[1] = float("nan")
        cont = contingency.Continuous(d, "sepal width")
        np.testing.assert_almost_equal(cont.col_unknowns, [0, 0, 0])
        np.testing.assert_almost_equal(
            cont.row_unknowns,
            [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.,
             0., 0., 0., 0., 0., 0., 0.])
        self.assertEqual(cont.unknowns, 1)

        # this one was failing before since the issue in _contingecy.pyx
        d.Y[:50] = np.zeros(50) * float("nan")
        cont = contingency.Continuous(d, "sepal width")
        np.testing.assert_almost_equal(cont.col_unknowns, [0, 0, 0])
        np.testing.assert_almost_equal(
            cont.row_unknowns,
            [0., 0., 1., 0., 0., 0., 0., 0., 1., 5., 5., 5., 2., 9., 6., 2.,
             3., 4., 2., 1., 1., 1., 1.])
        self.assertEqual(cont.unknowns, 1)

    @staticmethod
    def test_continuous_array_with_unknowns():
        """
        Test array_with_unknowns function
        """
        d = data.Table("iris")
        d.Y[:50] = np.zeros(50) * float("nan")
        cont = contingency.Continuous(d, "sepal width")
        correct_row_unknowns = [0., 0., 1., 0., 0., 0., 0., 0., 1., 6., 5., 5.,
                                2., 9., 6., 2., 3., 4., 2., 1., 1., 1., 1.]
        correct_row_unknowns_no_zero = [
            c for c in correct_row_unknowns if c > 0]
        correct_values_no_zero = [
            v for v, c in zip(cont.values, correct_row_unknowns) if c > 0]

        np.testing.assert_almost_equal(cont.row_unknowns, correct_row_unknowns)
        arr_unknowns = cont.array_with_unknowns
        np.testing.assert_almost_equal(
            arr_unknowns[-1][1], correct_row_unknowns_no_zero)
        np.testing.assert_almost_equal(
            arr_unknowns[-1][0], correct_values_no_zero)

        # check if other match to what we get with __getitem__
        for v1, v2 in zip(arr_unknowns[:-1], cont):
            np.testing.assert_almost_equal(v1, v2)

    def test_mixedtype_metas(self):
        import Orange
        zoo = Orange.data.Table("zoo")
        dom = Orange.data.Domain(zoo.domain.attributes[2:], zoo.domain.class_var,
                                 zoo.domain.metas + zoo.domain.attributes[:2])
        t = zoo.transform(dom)
        cont = contingency.get_contingency(zoo, 2, t.domain.metas[1])
        assert_dist_equal(cont["1"], [38, 5])
        assert_dist_equal(cont, [[4, 54], [38, 5]])
        zoo[25][t.domain.metas[1]] = float("nan")
        zoo[0][2] = float("nan")
        cont = contingency.get_contingency(zoo, 2, t.domain.metas[1])
        assert_dist_equal(cont["1"], [37, 5])
        assert_dist_equal(cont, [[4, 53], [37, 5]])
        np.testing.assert_almost_equal(cont.col_unknowns, [0, 1])
        np.testing.assert_almost_equal(cont.row_unknowns, [0, 1])
        self.assertEqual(0, cont.unknowns)

    @staticmethod
    def _construct_sparse():
        domain = data.Domain(
            [data.DiscreteVariable("d%i" % i, values=tuple("abc"))
             for i in range(10)] +
            [data.ContinuousVariable("c%i" % i) for i in range(10)],
            data.DiscreteVariable("y", values=tuple("abc")))

        #  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19
        #------------------------------------------------------------
        #     2     2  1  1  2        1           1  1     2  0  2
        #        1  1  0  0  1     2                 2     1  0
        #           1     2  0
        #
        #        2        0  1                   1.1
        #
        sdata = np.array([2, 2, 1, 1, 2, 1, 1, 1, 2, 0, 2,
                          1, 1, 0, 0, 1, 2, 2, 1, 0,
                          1, 2, 0,
                          2, 0, 1, 1.1])
        indices = [1, 3, 4, 5, 6, 9, 13, 14, 16, 17, 18,
                   2, 3, 4, 5, 6, 8, 14, 16, 17,
                   3, 5, 6,
                   2, 5, 6, 13]
        indptr = [0, 11, 20, 23, 23, 27]
        X = sp.csr_matrix((sdata, indices, indptr), shape=(5, 20))
        Y = np.array([[1, 2, 1, 0, 0]]).T
        return data.Table.from_numpy(domain, X, Y)

    def test_sparse(self):
        d = self._construct_sparse()
        cont = contingency.Discrete(d, 5)
        assert_dist_equal(cont[0], [2, 0, 0])
        assert_dist_equal(cont["b"], [0, 1, 1])
        assert_dist_equal(cont[2], [1, 0, 0])

        cont = contingency.Continuous(d, 14)
        assert_dist_equal(cont[0], [[], []])
        assert_dist_equal(cont["b"], [[1], [1]])
        assert_dist_equal(cont[2], [[2], [1]])

        cont = contingency.Continuous(d, "c3")
        assert_dist_equal(cont[0], [[1.1], [1]])
        assert_dist_equal(cont["b"], [[1], [1]])
        assert_dist_equal(cont[2], [[], []])

        d[4].set_class(1)
        cont = contingency.Continuous(d, 13)
        assert_dist_equal(cont[0], [[], []])
        assert_dist_equal(cont["b"], [[1, 1.1], [1, 1]])
        assert_dist_equal(cont[2], [[], []])

        cont = contingency.Continuous(d, 12)
        assert_dist_equal(cont[0], [[], []])
        assert_dist_equal(cont["b"], [[], []])
        assert_dist_equal(cont[2], [[], []])


    def test_get_contingency(self):
        d = self._construct_sparse()
        cont = contingency.get_contingency(d, 5)
        self.assertIsInstance(cont, contingency.Discrete)
        assert_dist_equal(cont[0], [2, 0, 0])
        assert_dist_equal(cont["b"], [0, 1, 1])
        assert_dist_equal(cont[2], [1, 0, 0])

        cont = contingency.get_contingency(d, "c4")
        self.assertIsInstance(cont, contingency.Continuous)
        assert_dist_equal(cont[0], [[], []])
        assert_dist_equal(cont["b"], [[1], [1]])
        assert_dist_equal(cont[2], [[2], [1]])

        cont = contingency.get_contingency(d, d.domain[13])
        self.assertIsInstance(cont, contingency.Continuous)
        assert_dist_equal(cont[0], [[1.1], [1]])
        assert_dist_equal(cont["b"], [[1], [1]])
        assert_dist_equal(cont[2], [[], []])
        assert_dist_equal(cont[2], [[], []])

    def test_get_contingencies(self):
        d = self._construct_sparse()
        conts = contingency.get_contingencies(d)

        self.assertEqual(len(conts), 20)

        cont = conts[5]
        self.assertIsInstance(cont, contingency.Discrete)
        assert_dist_equal(cont[0], [2, 0, 0])
        assert_dist_equal(cont["b"], [0, 1, 1])
        assert_dist_equal(cont[2], [1, 0, 0])

        cont = conts[14]
        self.assertIsInstance(cont, contingency.Continuous)
        assert_dist_equal(cont[0], [[], []])
        assert_dist_equal(cont["b"], [[1], [1]])
        assert_dist_equal(cont[2], [[2], [1]])

        conts = contingency.get_contingencies(d, skip_discrete=True)
        self.assertEqual(len(conts), 10)
        cont = conts[4]
        self.assertIsInstance(cont, contingency.Continuous)
        assert_dist_equal(cont[0], [[], []])
        assert_dist_equal(cont["b"], [[1], [1]])
        assert_dist_equal(cont[2], [[2], [1]])

        conts = contingency.get_contingencies(d, skip_continuous=True)
        self.assertEqual(len(conts), 10)
        cont = conts[5]
        self.assertIsInstance(cont, contingency.Discrete)
        assert_dist_equal(cont[0], [2, 0, 0])
        assert_dist_equal(cont["b"], [0, 1, 1])
        assert_dist_equal(cont[2], [1, 0, 0])

    def test_compute_contingency_metas(self):
        var1, var2 = self.test9.domain[-2], self.test9.domain[-4]
        cont = contingency.Discrete(self.test9, var1, var2)
        assert_dist_equal(cont, [[3, 0, 0], [0, 2, 0],
                                 [0, 0, 2], [0, 1, 0]])

    def test_compute_contingency_row_attribute_sparse(self):
        """
        Testing with sparse row variable since currently we do not test the
        situation when a row variable is sparse.
        """
        d = self.test9
        # make X sparse
        d.X = csr_matrix(d.X)
        var1, var2 = d.domain[0], d.domain[1]
        cont = contingency.Discrete(d, var1, var2)
        assert_dist_equal(cont, [[1, 0], [1, 0], [1, 0], [1, 0],
                                 [0, 1], [0, 1], [0, 1], [0, 1]])
        cont = contingency.Discrete(d, var2, var1)
        assert_dist_equal(cont, [[1, 1, 1, 1, 0, 0, 0, 0],
                                 [0, 0, 0, 0, 1, 1, 1, 1]])

        d.X = csc_matrix(d.X)
        cont = contingency.Discrete(d, var1, var2)
        assert_dist_equal(cont, [[1, 0], [1, 0], [1, 0], [1, 0],
                                 [0, 1], [0, 1], [0, 1], [0, 1]])
        cont = contingency.Discrete(d, var2, var1)
        assert_dist_equal(cont, [[1, 1, 1, 1, 0, 0, 0, 0],
                                 [0, 0, 0, 0, 1, 1, 1, 1]])

    def test_compute_contingency_invalid(self):
        rstate = np.random.RandomState(0xFFFF)
        X = data.ContinuousVariable("X")
        C = data.DiscreteVariable("C", values=["C{}".format(i + 1) for i in range(1024)])
        domain = data.Domain([X], [C])
        d = data.Table.from_numpy(
            domain,
            rstate.uniform(size=(20, 1)).round(1),
            rstate.randint(0, 1024, size=(20, 1)),
        )
        c = contingency.get_contingency(d, X, C)
        self.assertEqual(c.counts.shape[0], 1024)

        d.Y[5] = 1024
        with self.assertRaises(IndexError):
            contingency.get_contingency(d, X, C)

    def test_incompatible_arguments(self):
        """
        When providing data table unknowns should not be provided
        """
        self.assertRaises(
            TypeError, contingency.Discrete, self.zoo, 0, unknowns=0)
        self.assertRaises(
            TypeError, contingency.Discrete, self.zoo, 0, row_unknowns=0)
        self.assertRaises(
            TypeError, contingency.Discrete, self.zoo, 0, col_unknowns=0)

        self.assertRaises(
            TypeError, contingency.Continuous, self.zoo, 0, unknowns=0)
        self.assertRaises(
            TypeError, contingency.Continuous, self.zoo, 0, row_unknowns=0)
        self.assertRaises(
            TypeError, contingency.Continuous, self.zoo, 0, col_unknowns=0)

        # data with no class
        zoo_ = Table.from_table(Domain(self.zoo.domain.attributes), self.zoo)
        self.assertRaises(ValueError, contingency.Discrete, zoo_, 0)
        self.assertRaises(ValueError, contingency.Continuous, zoo_, 0)



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

from unittest import TestCase
import unittest
import pickle

import numpy as np
import scipy
import scipy.spatial
import scipy.stats
from scipy.sparse import csr_matrix

from Orange.data import (Table, Domain, ContinuousVariable,
                         DiscreteVariable, StringVariable, Instance)
from Orange.distance import (Euclidean, SpearmanR, SpearmanRAbsolute,
                             PearsonR, PearsonRAbsolute, Manhattan, Cosine,
                             Jaccard, _preprocess, MahalanobisDistance,
                             Bhattacharyya)
from Orange.distance.distance import _spearmanr2, _corrcoef2
from Orange.misc import DistMatrix
from Orange.tests import named_file, test_filename
from Orange.util import OrangeDeprecationWarning


def tables_equal(tab1, tab2):
    # TODO: introduce Table.__eq__() ???
    return (tab1 == tab2 or  # catches None
            np.all([i == j for i, j in zip(tab1, tab2)]))


class TestDistMatrix(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.dist = Euclidean(cls.iris)

    # pylint: disable=unsubscriptable-object
    def test_submatrix(self):
        sub = self.dist.submatrix([2, 3, 4])
        np.testing.assert_equal(sub, self.dist[2:5, 2:5])
        self.assertTrue(tables_equal(sub.row_items, self.dist.row_items[2:5]))

    def test_pickling(self):
        unpickled_dist = pickle.loads(pickle.dumps(self.dist))
        np.testing.assert_equal(unpickled_dist, self.dist)
        self.assertTrue(tables_equal(unpickled_dist.row_items,
                                     self.dist.row_items))
        self.assertTrue(tables_equal(unpickled_dist.col_items,
                                     self.dist.col_items))
        self.assertEqual(unpickled_dist.axis, self.dist.axis)

    def test_deprecated(self):
        a9 = np.arange(9).reshape(3, 3)
        m = DistMatrix(a9)
        with self.assertWarns(OrangeDeprecationWarning):
            self.assertEqual(m.dim, 3)
        with self.assertWarns(OrangeDeprecationWarning):
            np.testing.assert_almost_equal(m.X, a9)

    def test_from_file(self):
        with named_file(
                """3 axis=0 asymmetric col_labels row_labels
                    ann	bert	chad
                    danny	0.12	3.45	6.78
                    eve	9.01	2.34	5.67
                    frank	8.90	1.23	4.56""") as name:
            m = DistMatrix.from_file(name)
            np.testing.assert_almost_equal(m, np.array([[0.12, 3.45, 6.78],
                                                        [9.01, 2.34, 5.67],
                                                        [8.90, 1.23, 4.56]]))
            self.assertIsInstance(m.row_items, Table)
            self.assertIsInstance(m.col_items, Table)
            self.assertEqual([e.metas[0] for e in m.col_items],
                             ["ann", "bert", "chad"])
            self.assertEqual([e.metas[0] for e in m.row_items],
                             ["danny", "eve", "frank"])
            self.assertEqual(m.axis, 0)

        with named_file(
                """3 axis=1 row_labels
                    danny	0.12	3.45	6.78
                    eve 	9.01	2.34	5.67
                    frank	8.90""") as name:
            m = DistMatrix.from_file(name)
            np.testing.assert_almost_equal(m, np.array([[0.12, 9.01, 8.90],
                                                        [9.01, 2.34, 0],
                                                        [8.90, 0, 0]]))
            self.assertIsInstance(m.row_items, Table)
            self.assertIsNone(m.col_items)
            self.assertEqual([e.metas[0] for e in m.row_items],
                             ["danny", "eve", "frank"])
            self.assertEqual(m.axis, 1)

        with named_file(
                """3 axis=1 symmetric
                    0.12	3.45	6.78
                    9.01	2.34	5.67
                    8.90""") as name:
            m = DistMatrix.from_file(name)
        np.testing.assert_almost_equal(m, np.array([[0.12, 9.01, 8.90],
                                                    [9.01, 2.34, 0],
                                                    [8.90, 0, 0]]))

        with named_file(
                """3 row_labels
                    starič	0.12	3.45	6.78
                    aleš	9.01	2.34	5.67
                    anže	8.90""", encoding="utf-8""") as name:
            m = DistMatrix.from_file(name)
            np.testing.assert_almost_equal(m, np.array([[0.12, 9.01, 8.90],
                                                        [9.01, 2.34, 0],
                                                        [8.90, 0, 0]]))
            self.assertIsInstance(m.row_items, Table)
            self.assertIsNone(m.col_items)
            self.assertEqual([e.metas[0] for e in m.row_items],
                             ["starič", "aleš", "anže"])
            self.assertEqual(m.axis, 1)

        def assertErrorMsg(content, msg):
            with named_file(content) as name:
                with self.assertRaises(ValueError) as cm:
                    DistMatrix.from_file(name)
                self.assertEqual(str(cm.exception), msg)

        assertErrorMsg("",
                       "empty file")
        assertErrorMsg("axis=1\n1\t3\n4",
                       "distance file must begin with dimension")
        assertErrorMsg("3 col_labels\na\tb\n1\n\2\n3",
                       "mismatching number of column labels")
        assertErrorMsg("3 col_labels\na\tb\tc\td\n1\n\2\n3",
                       "mismatching number of column labels")
        assertErrorMsg("2\n  1\t2\t3\n  5",
                       "too many columns in matrix row 1")
        assertErrorMsg("2 row_labels\na\t1\t2\t3\nb\t5",
                       "too many columns in matrix row 'a'")
        assertErrorMsg("2 noflag\n  1\t2\t3\n  5",
                       "invalid flag 'noflag'")
        assertErrorMsg("2 noflag=5\n  1\t2\t3\n  5",
                       "invalid flag 'noflag=5'")
        assertErrorMsg("2\n1\n2\n3",
                       "too many rows")
        assertErrorMsg("2\n1\nasd",
                       "invalid element at row 2, column 1")
        assertErrorMsg("2 row_labels\na\t1\nb\tasd",
                       "invalid element at row 'b', column 1")
        assertErrorMsg("2 col_labels row_labels\nd\te\na\t1\nb\tasd",
                       "invalid element at row 'b', column 'd'")
        assertErrorMsg("2 col_labels\nd\te\n1\nasd",
                       "invalid element at row 2, column 'd'")

    def test_save(self):
        with named_file(
                """3 axis=1 row_labels
                    danny	0.12	3.45	6.78
                    eve 	9.01	2.34	5.67
                    frank	8.90""") as name:
            m = DistMatrix.from_file(name)
            m.save(name)
            m = DistMatrix.from_file(name)
            np.testing.assert_almost_equal(m, np.array([[0.12, 9.01, 8.90],
                                                        [9.01, 2.34, 0],
                                                        [8.90, 0, 0]]))
            self.assertIsInstance(m.row_items, Table)
            self.assertIsNone(m.col_items)
            self.assertEqual([e.metas[0] for e in m.row_items],
                             ["danny", "eve", "frank"])
            self.assertEqual(m.axis, 1)

        with named_file(
                """3 axis=0 asymmetric col_labels row_labels
                             ann	bert	chad
                    danny	0.12	3.45	6.78
                      eve	9.01	2.34	5.67
                    frank	8.90	1.23	4.56""") as name:
            m = DistMatrix.from_file(name)
            m.save(name)
            m = DistMatrix.from_file(name)
            np.testing.assert_almost_equal(m, np.array([[0.12, 3.45, 6.78],
                                                        [9.01, 2.34, 5.67],
                                                        [8.90, 1.23, 4.56]]))
            self.assertIsInstance(m.row_items, Table)
            self.assertIsInstance(m.col_items, Table)
            self.assertEqual([e.metas[0] for e in m.col_items],
                             ["ann", "bert", "chad"])
            self.assertEqual([e.metas[0] for e in m.row_items],
                             ["danny", "eve", "frank"])
            self.assertEqual(m.axis, 0)

    def test_numpy_type(self):
        """GH-3658"""
        data1 = np.array([1, 2], dtype=np.int64)
        data2 = np.array([2, 3], dtype=np.int64)
        dm1, dm2 = DistMatrix(data1), DistMatrix(data2)

        self.assertIsInstance(dm1.max(), np.int64)
        self.assertNotIsInstance(dm1.max(), int)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(dm1, dm2)


# noinspection PyTypeChecker
class TestEuclidean(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.sparse = Table.from_numpy(
            None, csr_matrix([[1, 2, 0], [0, 0, 3], [4, 0, 5]]))
        cls.dist = Euclidean

    def test_euclidean_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[1]),
            np.array([[0.53851648071346281]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[1], axis=1),
            np.array([[0.53851648071346281]]))

    def test_euclidean_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2]),
            np.array([[0., 0.53851648],
                      [0.53851648, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], axis=0),
            np.array([[0., 2.48394847, 5.09313263, 6.78969808],
                      [2.48394847, 0., 2.64007576, 4.327817],
                      [5.09313263, 2.64007576, 0., 1.69705627],
                      [6.78969808, 4.327817, 1.69705627, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[2], self.iris[:3]),
            np. array([[0.50990195, 0.3, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], self.iris[3]),
            np.array([[0.64807407],
                      [0.33166248]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], self.iris[:3]),
            np.array([[0., 0.53851648, 0.50990195],
                      [0.53851648, 0., 0.3]]))

    def test_euclidean_distance_sparse(self):
        np.testing.assert_almost_equal(
            self.dist(self.sparse),
            np.array([[0., 3.74165739, 6.164414],
                      [3.74165739, 0., 4.47213595],
                      [6.164414, 4.47213595, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.sparse, axis=0),
            np.array([[0., 4.12310563, 3.31662479],
                      [4.12310563, 0., 6.164414],
                      [3.31662479, 6.164414, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.sparse[:2]),
            np.array([[0., 3.74165739],
                      [3.74165739, 0.]]))

    def test_euclidean_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X),
            np.array([[0., 0.53851648],
                      [0.53851648, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[2].x, self.iris[:3].X),
            np. array([[0.50990195, 0.3, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X, self.iris[3].x),
            np.array([[0.64807407],
                      [0.33166248]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X, self.iris[:3].X),
            np.array([[0., 0.53851648, 0.50990195],
                      [0.53851648, 0., 0.3]]))


# noinspection PyTypeChecker
class TestManhattan(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.sparse = Table.from_numpy(
            None, csr_matrix([[1, 2, 0], [0, 0, 3], [4, 0, 5]]))
        cls.dist = Manhattan

    def test_manhattan_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[1]),
            np.array([[0.7]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[1], axis=1),
            np.array([[0.7]]))

    def test_manhattan_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2]),
            np.array([[0., 0.7],
                      [0.7, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], axis=0),
            np.array([[0., 3.5, 7.2, 9.6],
                      [3.5, 0., 3.7, 6.1],
                      [7.2, 3.7, 0., 2.4],
                      [9.6, 6.1, 2.4, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[2], self.iris[:3]),
            np.array([[0.8, 0.5, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], self.iris[3]),
            np.array([[1.],
                      [0.5]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], self.iris[:3]),
            np.array([[0., 0.7, 0.8],
                      [0.7, 0., 0.5]]))

    def test_manhattan_distance_sparse(self):
        np.testing.assert_almost_equal(
            self.dist(self.sparse),
            np.array([[0., 6., 10.],
                      [6., 0., 6.],
                      [10., 6., 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.sparse, axis=0),
            np.array([[0., 5., 5.],
                      [5., 0., 10.],
                      [5., 10., 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.sparse[:2]),
            np.array([[0., 6.],
                      [6., 0.]]))

    def test_manhattan_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[0].x, self.iris[1].x, axis=1),
            np.array([[0.7]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X),
            np.array([[0., 0.7],
                      [0.7, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[2].x, self.iris[:3].X),
            np.array([[0.8, 0.5, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X, self.iris[3].x),
            np.array([[1.],
                      [0.5]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X, self.iris[:3].X),
            np.array([[0., 0.7, 0.8],
                      [0.7, 0., 0.5]]))


# noinspection PyTypeChecker
class TestCosine(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.sparse = Table.from_numpy(
            None, csr_matrix([[1, 2, 0], [0, 0, 3], [4, 0, 5]]))
        cls.dist = Cosine

    def test_cosine_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[0]), np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[1]),
            np.array([[0.00142084]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[0], self.iris[1], axis=1),
            np.array([[0.00142084]]))

    def test_cosine_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2]),
            np.array([[0., 1.42083650e-03],
                      [1.42083650e-03, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], axis=0),
            np.array([[0.0, 1.61124231e-03, 1.99940020e-04, 1.99940020e-04],
                      [1.61124231e-03, 0.0, 2.94551450e-03, 2.94551450e-03],
                      [1.99940020e-04, 2.94551450e-03, 0.0, 0.0],
                      [1.99940020e-04, 2.94551450e-03, 0.0, 0.0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[2], self.iris[:3]),
            np.array([[1.26527175e-05, 1.20854727e-03, 0.0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], self.iris[3]),
            np.array([[0.00089939],
                      [0.00120607]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2], self.iris[:3]),
            np.array([[0.0, 1.42083650e-03, 1.26527175e-05],
                      [1.42083650e-03, 0.0, 1.20854727e-03]]))

    def test_cosine_distance_sparse(self):
        np.testing.assert_almost_equal(
            self.dist(self.sparse),
            np.array([[0.0, 1.00000000e+00, 7.20627882e-01],
                      [1.00000000e+00, 0.0, 2.19131191e-01],
                      [7.20627882e-01, 2.19131191e-01, 0.0]]))
        np.testing.assert_almost_equal(
            self.dist(self.sparse, axis=0),
            np.array([[0.0, 7.57464375e-01, 1.68109669e-01],
                      [7.57464375e-01, 0.0, 1.00000000e+00],
                      [1.68109669e-01, 1.00000000e+00, 0.0]]))
        np.testing.assert_almost_equal(
            self.dist(self.sparse[:2]),
            np.array([[0.0, 1.00000000e+00],
                      [1.00000000e+00, 0.0]]))

    def test_cosine_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.iris[0].x, self.iris[1].x, axis=1),
            np.array([[0.00142084]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X),
            np.array([[0., 1.42083650e-03],
                      [1.42083650e-03, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[2].x, self.iris[:3].X),
            np.array([[1.26527175e-05, 1.20854727e-03, 0.0]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X, self.iris[3].x),
            np.array([[0.00089939],
                      [0.00120607]]))
        np.testing.assert_almost_equal(
            self.dist(self.iris[:2].X, self.iris[:3].X),
            np.array([[0.0, 1.42083650e-03, 1.26527175e-05],
                      [1.42083650e-03, 0.0, 1.20854727e-03]]))


# noinspection PyTypeChecker
class TestJaccard(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.titanic = Table('titanic')[173:177]
        cls.dist = Jaccard

    def test_jaccard_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.titanic[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[0], self.titanic[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[0], self.titanic[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[0], self.titanic[2]),
            np.array([[0.5]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[0], self.titanic[2], axis=1),
            np.array([[0.5]]))

    def test_jaccard_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.titanic),
            np.array([[0., 0., 0.5, 0.5],
                      [0., 0., 0.5, 0.5],
                      [0.5, 0.5, 0., 0.],
                      [0.5, 0.5, 0., 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic, axis=0),
            np.array([[0., 1., 0.5],
                      [1., 0., 1.],
                      [0.5, 1., 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[2], self.titanic[:3]),
            np.array([[0.5, 0.5, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[:2], self.titanic[3]),
            np.array([[0.5],
                      [0.5]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[:2], self.titanic[:3]),
            np.array([[0., 0., 0.5],
                      [0., 0., 0.5]]))

    def test_jaccard_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.titanic[0].x, self.titanic[2].x, axis=1),
            np.array([[0.5]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic.X),
            np.array([[0., 0., 0.5, 0.5],
                      [0., 0., 0.5, 0.5],
                      [0.5, 0.5, 0., 0.],
                      [0.5, 0.5, 0., 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[2].x, self.titanic[:3].X),
            np.array([[0.5, 0.5, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[:2].X, self.titanic[3].x),
            np.array([[0.5],
                      [0.5]]))
        np.testing.assert_almost_equal(
            self.dist(self.titanic[:2].X, self.titanic[:3].X),
            np.array([[0., 0., 0.5],
                      [0., 0., 0.5]]))


# noinspection PyTypeChecker
class TestSpearmanR(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.breast = Table(test_filename(
            "datasets/breast-cancer-wisconsin.tab"))
        cls.dist = SpearmanR

    def test_spearmanr_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1]),
            np.array([[0.5083333333333333]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1], axis=1),
            np.array([[0.5083333333333333]]))

    def test_spearmanr_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2]),
            np.array([[0., 0.5083333333333333],
                      [0.5083333333333333, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:4]),
            np.array([[0., 0.50833333, 0.075, 0.61666667],
                      [0.50833333, 0., 0.38333333, 0.53333333],
                      [0.075, 0.38333333, 0., 0.63333333],
                      [0.61666667, 0.53333333, 0.63333333, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], axis=0),
            np.array([[0., 0.25, 0., 0.25, 0.25, 0.25, 0.75, 0.25, 0.25],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.75],
                      [0., 0.25, 0., 0.25, 0.25, 0.25, 0.75, 0.25, 0.25],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.75],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.75],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.75],
                      [0.75, 0.25, 0.75, 0.25, 0.25, 0.25, 0., 0.25, 1.],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.75],
                      [0.25, 0.75, 0.25, 0.75, 0.75, 0.75, 1., 0.75, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], self.breast[:4]),
            np.array([[0., 0.50833333, 0.075, 0.61666667],
                      [0.50833333, 0., 0.38333333, 0.53333333],
                      [0.075, 0.38333333, 0., 0.63333333]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[2], self.breast[:3]),
            np.array([[0.075, 0.3833333, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], self.breast[2]),
            np.array([[0.075],
                      [0.3833333],
                      [0.]]))

    def test_spearmanr_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0].x, self.breast[1].x, axis=1),
            np.array([[0.5083333333333333]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2].X),
            np.array([[0., 0.5083333333333333],
                      [0.5083333333333333, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[2].x, self.breast[:3].X),
            np.array([[0.075, 0.3833333, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3].X, self.breast[2].x),
            np.array([[0.075],
                      [0.3833333],
                      [0.]]))

    def test_spearmanr2(self):
        # Test that _spearnmanr2 returns the same result that stats.spearmanr
        # would
        n, m = tuple(np.random.randint(2, 5, size=2))
        mean = np.random.uniform(-1, 1, size=m)
        cov = np.random.uniform(0, 1./m, size=(m, m))
        cov = (cov + cov.T) / 2
        cov.flat[::m + 1] = 1.0
        X1 = np.random.multivariate_normal(mean, cov, size=n)
        X2 = np.random.multivariate_normal(mean, cov, size=n)
        expected = scipy.stats.spearmanr(X1, X2, axis=1)[0][:n, n:]
        np.testing.assert_almost_equal(
            _spearmanr2(X1, X2, axis=1),
            expected,
            decimal=9
        )

        expected = scipy.stats.spearmanr(X1, X2, axis=0)[0][:m, m:]
        np.testing.assert_almost_equal(
            _spearmanr2(X1, X2, axis=0),
            expected,
            decimal=9,
        )


# noinspection PyTypeChecker
class TestSpearmanRAbsolute(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.breast = Table(test_filename(
            "datasets/breast-cancer-wisconsin.tab"))
        cls.dist = SpearmanRAbsolute

    def test_spearmanrabsolute_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1]),
            np.array([[0.49166666666666664]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1], axis=1),
            np.array([[0.49166666666666664]]))

    def test_spearmanrabsolute_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2]),
            np.array([[0., 0.49166667],
                      [0.49166667, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], axis=0),
            np.array([[0., 0.25, 0., 0.25, 0.25, 0.25, 0.25, 0.25, 0.25],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.25],
                      [0., 0.25, 0., 0.25, 0.25, 0.25, 0.25, 0.25, 0.25],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.25],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.25],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.25],
                      [0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0., 0.25, 0.],
                      [0.25, 0., 0.25, 0., 0., 0., 0.25, 0., 0.25],
                      [0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0., 0.25, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], self.breast[:4]),
            np.array([[0., 0.49166667, 0.075, 0.38333333],
                      [0.49166667, 0., 0.38333333, 0.46666667],
                      [0.075, 0.38333333, 0., 0.36666667]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[3], self.breast[:4]),
            np.array([[0.3833333, 0.4666667, 0.3666667, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:4], self.breast[3]),
            np.array([[0.3833333],
                      [0.4666667],
                      [0.3666667],
                      [0.]]))

    def test_spearmanrabsolute_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0].x, self.breast[1].x, axis=1),
            np.array([[0.49166666666666664]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2].X),
            np.array([[0., 0.49166667],
                      [0.49166667, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[3].x, self.breast[:4].X),
            np.array([[0.3833333, 0.4666667, 0.3666667, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:4].X, self.breast[3].x),
            np.array([[0.3833333],
                      [0.4666667],
                      [0.3666667],
                      [0.]]))


# noinspection PyTypeChecker
class TestPearsonR(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.breast = Table(test_filename(
            "datasets/breast-cancer-wisconsin.tab"))
        cls.dist = PearsonR

    def test_pearsonr_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1]),
            np.array([[0.48462293898088876]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1], axis=1),
            np.array([[0.48462293898088876]]))

    def test_pearsonr_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2]),
            np.array([[0., 0.48462294],
                      [0.48462294, 0.]]))
        # pylint: disable=line-too-long
        # Because it looks better
        np.testing.assert_almost_equal(
            self.dist(self.breast[:20], axis=0),
            np.array([[0., 0.10239274, 0.12786763, 0.13435117, 0.15580385, 0.27429811, 0.21006195, 0.24072005, 0.42847752],
                      [0.10239274, 0., 0.01695375, 0.10313851, 0.1138925, 0.16978203, 0.1155948, 0.08043531, 0.43326547],
                      [0.12786763, 0.01695375, 0., 0.16049178, 0.13692762, 0.21784201, 0.11607395, 0.06493949, 0.46590168],
                      [0.13435117, 0.10313851, 0.16049178, 0., 0.07181648, 0.15585667, 0.13891172, 0.21622332, 0.37404826],
                      [0.15580385, 0.1138925, 0.13692762, 0.07181648, 0., 0.16301705, 0.17324382, 0.21452448, 0.42283252],
                      [0.27429811, 0.16978203, 0.21784201, 0.15585667, 0.16301705, 0., 0.25512861, 0.29560909, 0.42766076],
                      [0.21006195, 0.1155948, 0.11607395, 0.13891172, 0.17324382, 0.25512861, 0., 0.14419442, 0.57976119],
                      [0.24072005, 0.08043531, 0.06493949, 0.21622332, 0.21452448, 0.29560909, 0.14419442, 0., 0.45930368],
                      [0.42847752, 0.43326547, 0.46590168, 0.37404826, 0.42283252, 0.42766076, 0.57976119, 0.45930368, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], self.breast[:4]),
            np.array([[0., 0.48462294, 0.10133593, 0.5016744],
                      [0.48462294, 0., 0.32783865, 0.57317387],
                      [0.10133593, 0.32783865, 0., 0.63789635]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[2], self.breast[:3]),
            np.array([[0.10133593, 0.32783865, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], self.breast[2]),
            np.array([[0.10133593],
                      [0.32783865],
                      [0.]]))

    def test_pearsonr_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0].x, self.breast[1].x, axis=1),
            np.array([[0.48462293898088876]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2].X),
            np.array([[0., 0.48462294],
                      [0.48462294, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[2].x, self.breast[:3].X),
            np.array([[0.10133593, 0.32783865, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3].X, self.breast[2].x),
            np.array([[0.10133593],
                      [0.32783865],
                      [0.]]))

    def test_corrcoef2(self):
        # Test that _corrcoef2 returns the same result that np.corrcoef would
        n, m = tuple(np.random.randint(2, 5, size=2))
        mean = np.random.uniform(-1, 1, size=m)
        cov = np.random.uniform(0, 1./m, size=(m, m))
        cov = (cov + cov.T) / 2
        cov.flat[::m + 1] = 1.0
        X1 = np.random.multivariate_normal(mean, cov, size=n)
        X2 = np.random.multivariate_normal(mean, cov, size=n)
        expected = np.corrcoef(X1, X2, rowvar=True)[:n, n:]
        np.testing.assert_almost_equal(
            _corrcoef2(X1, X2, axis=1),
            expected,
            decimal=9
        )

        expected = np.corrcoef(X1, X2, rowvar=False)[:m, m:]
        np.testing.assert_almost_equal(
            _corrcoef2(X1, X2, axis=0),
            expected,
            decimal=9,
        )

        with self.assertRaises(ValueError):
            _corrcoef2(X1, X2, axis=10)


# noinspection PyTypeChecker
class TestPearsonRAbsolute(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.breast = Table(test_filename(
            "datasets/breast-cancer-wisconsin.tab"))
        cls.dist = PearsonRAbsolute

    def test_pearsonrabsolute_distance_one_example(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0]),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[0], axis=1),
            np.array([[0]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1]),
            np.array([[0.48462293898088876]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[0], self.breast[1], axis=1),
            np.array([[0.48462293898088876]]))

    def test_pearsonrabsolute_distance_many_examples(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2]),
            np.array([[0., 0.48462294],
                      [0.48462294, 0.]]))
        # pylint: disable=line-too-long
        # Because it looks better
        np.testing.assert_almost_equal(
            self.dist(self.breast[:20], axis=0),
            np.array([[0., 0.10239274, 0.12786763, 0.13435117, 0.15580385, 0.27429811, 0.21006195, 0.24072005, 0.42847752],
                      [0.10239274, 0., 0.01695375, 0.10313851, 0.1138925, 0.16978203, 0.1155948, 0.08043531, 0.43326547],
                      [0.12786763, 0.01695375, 0., 0.16049178, 0.13692762, 0.21784201, 0.11607395, 0.06493949, 0.46590168],
                      [0.13435117, 0.10313851, 0.16049178, 0., 0.07181648, 0.15585667, 0.13891172, 0.21622332, 0.37404826],
                      [0.15580385, 0.1138925, 0.13692762, 0.07181648, 0., 0.16301705, 0.17324382, 0.21452448, 0.42283252],
                      [0.27429811, 0.16978203, 0.21784201, 0.15585667, 0.16301705, 0., 0.25512861, 0.29560909, 0.42766076],
                      [0.21006195, 0.1155948, 0.11607395, 0.13891172, 0.17324382, 0.25512861, 0., 0.14419442, 0.42023881],
                      [0.24072005, 0.08043531, 0.06493949, 0.21622332, 0.21452448, 0.29560909, 0.14419442, 0., 0.45930368],
                      [0.42847752, 0.43326547, 0.46590168, 0.37404826, 0.42283252, 0.42766076, 0.42023881, 0.45930368, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:3], self.breast[:4]),
            np.array([[0., 0.48462294, 0.10133593, 0.4983256],
                      [0.48462294, 0., 0.32783865, 0.42682613],
                      [0.10133593, 0.32783865, 0., 0.36210365]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[2], self.breast[:3]),
            np.array([[0.10133593, 0.32783865, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2], self.breast[3]),
            np.array([[0.4983256],
                      [0.42682613]]))

    def test_pearsonrabsolute_distance_numpy(self):
        np.testing.assert_almost_equal(
            self.dist(self.breast[0].x, self.breast[1].x, axis=1),
            np.array([[0.48462293898088876]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2].X),
            np.array([[0., 0.48462294],
                      [0.48462294, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[2].x, self.breast[:3].X),
            np.array([[0.10133593, 0.32783865, 0.]]))
        np.testing.assert_almost_equal(
            self.dist(self.breast[:2].X, self.breast[3].x),
            np.array([[0.4983256],
                      [0.42682613]]))


# noinspection PyTypeChecker
# Pylint doesn't get magic __new__ operators
# pylint: disable=not-callable
class TestMahalanobis(TestCase):
    def setUp(self):
        self.n, self.m = 10, 5
        self.x = np.random.rand(self.n, self.m)
        self.x1 = np.random.rand(self.m)
        self.x2 = np.random.rand(self.m)

    def test_correctness(self):
        mah = MahalanobisDistance(self.x)
        d = scipy.spatial.distance.pdist(self.x, 'mahalanobis')
        d = scipy.spatial.distance.squareform(d)
        for i in range(self.n):
            for j in range(self.n):
                self.assertAlmostEqual(d[i][j], mah(self.x[i], self.x[j]),
                                       delta=1e-5)

    def test_attributes(self):
        metric = MahalanobisDistance(self.x)
        self.assertEqual(metric(self.x[0], self.x[1]).shape, (1, 1))
        self.assertEqual(metric(self.x).shape, (self.n, self.n))
        self.assertEqual(metric(self.x[0:3], self.x[5:7]).shape, (3, 2))
        self.assertEqual(metric(self.x1, self.x2).shape, (1, 1))
        metric(self.x, impute=True)
        metric(self.x[:-1, :])
        self.assertRaises(ValueError, metric, self.x[:, :-1])
        self.assertRaises(ValueError, metric, self.x1[:-1], self.x2)
        self.assertRaises(ValueError, metric, self.x1, self.x2[:-1])
        self.assertRaises(ValueError, metric, self.x.T)

    def test_iris(self):
        tab = Table('iris')
        metric = MahalanobisDistance(tab)
        self.assertEqual(metric(tab).shape, (150, 150))
        self.assertEqual(metric(tab[0], tab[1]).shape, (1, 1))

    def test_dimensions(self):
        x = Table('iris')[:20].X
        xt = Table('iris')[:20].X.T
        mah = MahalanobisDistance(x)
        mah(x[0], x[1])
        mah = MahalanobisDistance(xt)
        mah(xt[0], xt[1])


class TestBhattacharyya(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dist = Bhattacharyya

    def test_dense_array(self):
        #Also checks normalization
        data = Table('iris')
        true_out = np.array([[0, 4.48049499e-04, 2.07117086e-05],
                             [4.48049499e-04, 0, 3.65052724e-04],
                             [2.07117086e-05, 3.65052724e-04, 0]])
        np.testing.assert_array_almost_equal(self.dist(data.X[:3]), true_out)

    def test_sparse_array(self):
        data = csr_matrix([[0.5, 0.5], [0, 0.5]])
        self.assertAlmostEqual(self.dist(data[0], data[1]), 0.3465735902799726, delta=1e-5)

    def test_columns(self):
        data = np.array([[0.5, 0.2], [0.5, 0.8]])
        true_out = np.array([[0, 0.05268025782891318],
                             [0.05268025782891318, 0]])
        np.testing.assert_array_almost_equal(self.dist(data, axis=0), true_out)

    def test_negative_input(self):
        a = np.array([0, np.nan])
        b = np.array([1, 1])
        self.assertRaises(ValueError, self.dist, a, b)
        a[1] = -1
        self.assertRaises(ValueError, self.dist, a, b)
        a = csr_matrix(a)
        b = csr_matrix(b)
        self.assertRaises(ValueError, self.dist, a, b)


class TestDistances(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test5 = Table(test_filename('datasets/test5.tab'))

    def test_preprocess(self):
        domain = Domain([ContinuousVariable("c"),
                         DiscreteVariable("d", values=('a', 'b'))],
                        [DiscreteVariable("cls", values=('e', 'f'))],
                        [StringVariable("m")])
        table = Table.from_list(domain, [[1, 'a', 'e', 'm1'],
                                         [2, 'b', 'f', 'm2']])
        new_table = _preprocess(table)
        np.testing.assert_equal(new_table.X, table.X[:, 0].reshape(2, 1))
        np.testing.assert_equal(new_table.Y, table.Y)
        np.testing.assert_equal(new_table.metas, table.metas)
        self.assertEqual([a.name for a in new_table.domain.attributes],
                         [a.name for a in table.domain.attributes
                          if a.is_continuous])
        self.assertEqual(new_table.domain.class_vars, table.domain.class_vars)
        self.assertEqual(new_table.domain.metas, table.domain.metas)

    def test_preprocess_multiclass(self):
        table = self.test5
        new_table = _preprocess(table)
        np.testing.assert_equal(new_table.Y, table.Y)
        self.assertEqual([a.name for a in new_table.domain.attributes],
                         [a.name for a in table.domain.attributes if
                          a.is_continuous and not all(np.isnan(table[:, a].X))])
        self.assertEqual(new_table.domain.class_vars, table.domain.class_vars)

    def test_preprocess_impute(self):
        new_table = _preprocess(self.test5)
        self.assertFalse(np.isnan(new_table.X).any())

    def test_distance_to_instance(self):
        iris = Table('iris')
        inst = Instance(iris.domain, np.concatenate((iris[1].x, iris[1].y)))
        self.assertEqual(Euclidean(iris[1], inst), 0)

if __name__ == '__main__':
    unittest.main()




from orangewidget.utils.filedialogs import (
    open_filename_dialog_save, open_filename_dialog,
    RecentPath, RecentPathsWidgetMixin, RecentPathsWComboMixin,
)
# imported for backcompatibility
from orangewidget.utils.filedialogs import (  # pylint: disable=unused-import
    fix_extension, format_filter, get_file_name, Compression
)

from Orange.data.io import FileFormat
from Orange.util import deprecated

__all__ = [
    "open_filename_dialog_save", "open_filename_dialog",
    "RecentPath", "RecentPathsWidgetMixin", "RecentPathsWComboMixin",
]


@deprecated
def dialog_formats():
    """
    Return readable file types for QFileDialogs.
    """
    return ("All readable files ({});;".format(
        '*' + ' *'.join(FileFormat.readers.keys())) +
            ";;".join("{} (*{})".format(f.DESCRIPTION, ' *'.join(f.EXTENSIONS))
                      for f in sorted(set(FileFormat.readers.values()),
                                      key=list(FileFormat.readers.values()).index)))



import numpy as np
import scipy.sparse as sp

from Orange.projection import _som


class SOM:
    def __init__(self, dim_x, dim_y,
                 hexagonal=False, pca_init=True, random_seed=None):
        self.dim_x = dim_x
        self.dim_y = dim_y
        self.weights = self.ssum_weights = None
        self.hexagonal = hexagonal
        self.pca_init = pca_init
        self.random_seed = random_seed

    def init_weights_random(self, x):
        random = (np.random if self.random_seed is None
                  else np.random.RandomState(self.random_seed))
        self.weights = random.rand(self.dim_y, self.dim_x, x.shape[1])
        norms = np.sum(self.weights ** 2, axis=2)
        norms[norms == 0] = 1
        self.weights /= norms[:, :, None]
        self.ssum_weights = np.ones((self.dim_y, self.dim_x))

    def init_weights_pca(self, x):
        pc_length, pc = np.linalg.eig(np.cov(x.T))
        c0, c1, *_ = np.argsort(pc_length)
        pc0, pc1 = np.real(pc[c0]), np.real(pc[c1])
        self.weights = np.empty((self.dim_y, self.dim_x, x.shape[1]))
        for i, c1 in enumerate(np.linspace(-1, 1, self.dim_y)):
            for j, c2 in enumerate(np.linspace(-1, 1, self.dim_x)):
                self.weights[i, j] = c1 * pc0 + c2 * pc1
        norms = np.sum(self.weights ** 2, axis=2)
        norms[norms == 0] = 1
        self.weights /= norms[:, :, None]
        self.ssum_weights = np.ones((self.dim_y, self.dim_x))

    def fit(self, x, n_iterations, learning_rate=0.5, sigma=1.0, callback=None):
        if sp.issparse(x):
            f = _som.update_sparse_hex if self.hexagonal else _som.update_sparse

            def update(decay):
                f(self.weights, self.ssum_weights, x,
                  sigma / decay, learning_rate / decay)
        else:
            f = _som.update_hex if self.hexagonal else _som.update

            def update(decay):
                f(self.weights, x,
                  sigma / decay, learning_rate / decay)

        if self.pca_init and not sp.issparse(x) and x.shape[1] > 1:
            self.init_weights_pca(x)
        else:
            self.init_weights_random(x)

        for iteration in range(n_iterations):
            update(1 + iteration / (n_iterations / 2))
            if callback is not None and not callback(iteration / n_iterations):
                break

    def winners(self, x):
        return self.winner_from_weights(
            x, self.weights, self.ssum_weights, self.hexagonal)

    @staticmethod
    def winner_from_weights(x, weights, ssum_weights, hexagonal):
        if sp.issparse(x):
            return _som.get_winners_sparse(
                weights, ssum_weights, x, int(hexagonal))
        else:
            return _som.get_winners(weights, x, int(hexagonal))



import unittest
from unittest.mock import patch, Mock, call

import numpy as np

from Orange.data import DiscreteVariable, ContinuousVariable, Domain, Table
from Orange.preprocess import Normalize
from Orange.projection import manifold, TSNE
from Orange.projection.manifold import TSNEModel
from Orange.widgets.tests.base import (
    WidgetTest, WidgetOutputsTestMixin, ProjectionWidgetTestMixin
)
from Orange.widgets.unsupervised.owtsne import OWtSNE, TSNERunner, Task, prepare_tsne_obj


class DummyTSNE(manifold.TSNE):
    def fit(self, X, Y=None):
        return np.ones((len(X), 2), float)


class DummyTSNEModel(manifold.TSNEModel):
    def transform(self, X, **kwargs):
        return np.ones((len(X), 2), float)

    def optimize(self, n_iter, **kwargs):
        return self


class TestOWtSNE(WidgetTest, ProjectionWidgetTestMixin, WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        WidgetOutputsTestMixin.init(cls)
        cls.same_input_output_domain = False

        cls.signal_name = "Data"
        cls.signal_data = cls.data

    def setUp(self):
        # For almost all the tests, we won't need to verify t-SNE validity and
        # the tests will run much faster if we dummy them out
        self.tsne = patch("Orange.projection.manifold.TSNE", new=DummyTSNE)
        self.tsne_model = patch("Orange.projection.manifold.TSNEModel", new=DummyTSNEModel)
        self.tsne.start()
        self.tsne_model.start()

        self.widget = self.create_widget(OWtSNE, stored_settings={"multiscale": False})

        self.class_var = DiscreteVariable("Stage name", values=("STG1", "STG2"))
        self.attributes = [ContinuousVariable("GeneName" + str(i)) for i in range(5)]
        self.domain = Domain(self.attributes, class_vars=self.class_var)
        self.empty_domain = Domain([], class_vars=self.class_var)

    def tearDown(self):
        # Some tests may not wait for the widget to finish, and the patched
        # methods might be unpatched before the widget finishes, resulting in
        # a very confusing crash.
        self.widget.onDeleteWidget()
        try:
            self.restore_mocked_functions()
        # If `restore_mocked_functions` was called in the test itself, stopping
        # the patchers here will raise a RuntimeError
        except RuntimeError as e:
            if str(e) != "stop called on unstarted patcher":
                raise e

    def restore_mocked_functions(self):
        self.tsne.stop()
        self.tsne_model.stop()

    def test_wrong_input(self):
        # no data
        data = None
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)

        # <2 rows
        data = Table.from_list(self.domain, [[1, 2, 3, 4, 5, 'STG1']])
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.not_enough_rows.is_shown())

        # no attributes
        data = Table.from_list(self.empty_domain, [['STG1']] * 2)
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.not_enough_cols.is_shown())

        # one attributes
        data = Table.from_list(self.empty_domain, [[1, 'STG1'],
                                                   [2, 'STG1']])
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.not_enough_cols.is_shown())

        # constant data
        data = Table.from_list(self.domain, [[1, 2, 3, 4, 5, 'STG1']] * 2)
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNone(self.widget.data)
        self.assertTrue(self.widget.Error.constant_data.is_shown())

        # correct input
        data = Table.from_list(self.domain, [[1, 2, 3, 4, 5, 'STG1'],
                                             [5, 4, 3, 2, 1, 'STG1']])
        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()
        self.assertIsNotNone(self.widget.data)
        self.assertFalse(self.widget.Error.not_enough_rows.is_shown())
        self.assertFalse(self.widget.Error.not_enough_cols.is_shown())
        self.assertFalse(self.widget.Error.constant_data.is_shown())

    def test_input(self):
        data = Table.from_list(self.domain, [[1, 1, 1, 1, 1, 'STG1'],
                                             [2, 2, 2, 2, 2, 'STG1'],
                                             [4, 4, 4, 4, 4, 'STG2'],
                                             [5, 5, 5, 5, 5, 'STG2']])

        self.send_signal(self.widget.Inputs.data, data)
        self.wait_until_stop_blocking()

    def test_attr_models(self):
        """Check possible values for 'Color', 'Shape', 'Size' and 'Label'"""
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        controls = self.widget.controls
        for var in self.data.domain.class_vars + self.data.domain.metas:
            self.assertIn(var, controls.attr_color.model())
            self.assertIn(var, controls.attr_label.model())
            if var.is_continuous:
                self.assertIn(var, controls.attr_size.model())
                self.assertNotIn(var, controls.attr_shape.model())
            if var.is_discrete:
                self.assertNotIn(var, controls.attr_size.model())
                self.assertIn(var, controls.attr_shape.model())

    def test_multiscale_changed_updates_ui(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertFalse(self.widget.controls.multiscale.isChecked())
        self.assertTrue(self.widget.perplexity_spin.isEnabled())
        self.widget.controls.multiscale.setChecked(True)
        self.assertFalse(self.widget.perplexity_spin.isEnabled())

        settings = self.widget.settingsHandler.pack_data(self.widget)
        w = self.create_widget(OWtSNE, stored_settings=settings)
        self.send_signal(w.Inputs.data, self.data, widget=w)
        self.assertTrue(w.controls.multiscale.isChecked())
        self.assertFalse(w.perplexity_spin.isEnabled())
        w.onDeleteWidget()

    def test_normalize_data(self):
        # Normalization should be checked by default
        self.assertTrue(self.widget.controls.normalize.isChecked())
        with patch("Orange.preprocess.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, self.data)
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            self.wait_until_finished()
            normalize.assert_called_once()

        # Disable checkbox
        self.widget.controls.normalize.setChecked(False)
        self.assertFalse(self.widget.controls.normalize.isChecked())
        with patch("Orange.preprocess.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, self.data)
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            self.wait_until_finished()
            normalize.assert_not_called()

        # Normalization shouldn't work on sparse data
        self.widget.controls.normalize.setChecked(True)
        self.assertTrue(self.widget.controls.normalize.isChecked())

        sparse_data = self.data.to_sparse()
        with patch("Orange.preprocess.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, sparse_data)
            self.assertFalse(self.widget.controls.normalize.isEnabled())
            self.wait_until_finished()
            normalize.assert_not_called()

    @patch("Orange.projection.manifold.TSNEModel.optimize")
    def test_exaggeration_is_passed_through_properly(self, optimize):
        def _check_exaggeration(call, exaggeration):
            # Check the last call to `optimize`, so we catch one during the
            # regular regime
            _, _, kwargs = call.mock_calls[-1]
            self.assertIn("exaggeration", kwargs)
            self.assertEqual(kwargs["exaggeration"], exaggeration)

        # Since optimize needs to return a valid TSNEModel instance and it is
        # impossible to return `self` in a mock, we'll prepare this one ahead
        # of time and use this
        optimize.return_value = DummyTSNE()(self.data)

        # Set value to 1
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.run_button.clicked.emit()  # stop initial run
        self.wait_until_stop_blocking()
        self.widget.controls.exaggeration.setValue(1)
        self.widget.run_button.clicked.emit()  # run with exaggeration 1
        self.wait_until_finished()
        _check_exaggeration(optimize, 1)

        # Reset and clear state
        self.send_signal(self.widget.Inputs.data, None)
        optimize.reset_mock()

        # Change to 3
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.run_button.clicked.emit()  # stop initial run
        self.wait_until_stop_blocking()
        self.widget.controls.exaggeration.setValue(3)
        self.widget.run_button.clicked.emit()  # run with exaggeration 1
        self.wait_until_finished()
        _check_exaggeration(optimize, 3)

    def test_plot_once(self):
        """Test if data is plotted only once but committed on every input change"""
        self.widget.setup_plot = Mock()
        self.widget.commit = self.widget.unconditional_commit = Mock()

        self.send_signal(self.widget.Inputs.data, self.data)
        # TODO: The base widget immediately calls `setup_plot` and `commit`
        # even though there's nothing to show yet. Unfortunately, fixing this
        # would require changing `OWDataProjectionWidget` in some strange way,
        # so as a temporary fix, we reset the mocks, so they reflect the calls
        # when the result was available.
        self.widget.setup_plot.reset_mock()
        self.widget.commit.reset_mock()
        self.wait_until_finished()

        self.widget.setup_plot.assert_called_once()
        self.widget.commit.assert_called_once()

        self.widget.commit.reset_mock()
        self.send_signal(self.widget.Inputs.data_subset, self.data[::10])
        self.wait_until_stop_blocking()

        self.widget.setup_plot.assert_called_once()
        self.widget.commit.assert_called_once()

    def test_modified_info_message_behaviour(self):
        """Information messages should be cleared if the data changes or if
        the data is set to None."""
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden by default"
        )

        self.widget.controls.multiscale.setChecked(False)
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden even after toggling "
            "options if no data is on input"
        )

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden after the widget "
            "computes the embedding"
        )

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be hidden when reloading the "
            "same data set and no previous messages were shown"
        )

        self.widget.controls.multiscale.setChecked(True)
        self.assertTrue(
            self.widget.Information.modified.is_shown(),
            "The modified info message should be shown when a setting is "
            "changed, but the embedding is not recomputed"
        )

        self.send_signal(self.widget.Inputs.data, Table("housing"))
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The information message was not cleared on new data"
        )
        # Flip; the info message should now be shown again.
        self.widget.controls.multiscale.setChecked(False)
        assert self.widget.Information.modified.is_shown()  # sanity check

        self.send_signal(self.widget.Inputs.data, None)
        self.wait_until_stop_blocking()
        self.assertFalse(
            self.widget.Information.modified.is_shown(),
            "The information message was not cleared on no data"
        )

    def test_invalidation_flow(self):
        # pylint: disable=protected-access
        w = self.widget
        # Setup widget: send data to input with global structure "off", then
        # set global structure "on" (after the embedding is computed)
        w.controls.multiscale.setChecked(False)
        self.send_signal(w.Inputs.data, self.data)
        self.wait_until_finished()
        self.assertFalse(self.widget.Information.modified.is_shown())
        # All the embedding components should computed
        self.assertIsNotNone(w.pca_projection)
        self.assertIsNotNone(w.affinities)
        self.assertIsNotNone(w.tsne_embedding)
        # All the invalidation flags should be set to false
        self.assertFalse(w._invalidated.pca_projection)
        self.assertFalse(w._invalidated.affinities)
        self.assertFalse(w._invalidated.tsne_embedding)

        # Trigger invalidation
        w.controls.multiscale.setChecked(True)
        self.assertTrue(self.widget.Information.modified.is_shown())
        # Setting `multiscale` to true should set the invalidate flags for
        # the affinities and embedding, but not the pca_projection
        self.assertFalse(w._invalidated.pca_projection)
        self.assertTrue(w._invalidated.affinities)
        self.assertTrue(w._invalidated.tsne_embedding)

        # The flags should now be set, but the embedding should still be
        # available when selecting a subset of data and such
        self.assertIsNotNone(w.pca_projection)
        self.assertIsNotNone(w.affinities)
        self.assertIsNotNone(w.tsne_embedding)

        # We should still be able to send a data subset to the input and have
        # the points be highlighted
        self.send_signal(w.Inputs.data_subset, self.data[:10])
        self.wait_until_stop_blocking()
        subset = [brush.color().name() == "#46befa" for brush in
                  w.graph.scatterplot_item.data["brush"][:10]]
        other = [brush.color().name() == "#000000" for brush in
                 w.graph.scatterplot_item.data["brush"][10:]]
        self.assertTrue(all(subset))
        self.assertTrue(all(other))

        # Clear the data subset
        self.send_signal(w.Inputs.data_subset, None)

        # Run the optimization
        self.widget.run_button.clicked.emit()
        self.wait_until_stop_blocking()
        # All of the inavalidation flags should have been cleared
        self.assertFalse(w._invalidated)


class TestTSNERunner(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = Table("iris")

    def test_run(self):
        state = Mock()
        state.is_interruption_requested = Mock(return_value=False)

        task = TSNERunner.run(Task(data=self.data, perplexity=30), state)

        self.assertEqual(len(state.set_status.mock_calls), 4)
        state.set_status.assert_has_calls([
            call("Computing PCA..."),
            call("Preparing initialization..."),
            call("Finding nearest neighbors..."),
            call("Running optimization..."),
        ])

        self.assertIsInstance(task.pca_projection, Table)
        self.assertIsInstance(task.tsne, TSNE)
        self.assertIsInstance(task.tsne_embedding, TSNEModel)

    def test_run_do_not_modify_model_inplace(self):
        state = Mock()
        state.is_interruption_requested.return_value = True

        task = Task(data=self.data, perplexity=30, multiscale=False, exaggeration=1)
        # Run through all the steps to prepare the t-SNE object
        task.tsne = prepare_tsne_obj(
            task.data, task.perplexity, task.multiscale, task.exaggeration
        )
        TSNERunner.compute_pca(task, state)
        TSNERunner.compute_initialization(task, state)
        TSNERunner.compute_affinities(task, state)
        # Run the t-SNE iteration once to create the object
        TSNERunner.compute_tsne(task, state)

        # Make sure that running t-SNE for another iteration returns a new object
        tsne_obj_before = task.tsne_embedding
        state.reset_mock()
        TSNERunner.compute_tsne(task, state)
        tsne_obj_after = task.tsne_embedding

        state.set_partial_result.assert_called_once()
        self.assertIsNot(tsne_obj_before, tsne_obj_after)


if __name__ == "__main__":
    unittest.main()



from itertools import chain
from copy import deepcopy

import numpy as np
import scipy.sparse as sp

from AnyQt.QtCore import Qt, QAbstractTableModel
from AnyQt.QtGui import QColor, QFont
from AnyQt.QtWidgets import QComboBox, QTableView, QSizePolicy

from Orange.data import DiscreteVariable, ContinuousVariable, StringVariable, \
    TimeVariable, Domain
from Orange.misc.collections import natural_sorted
from Orange.data.util import get_unique_names_duplicates
from Orange.statistics.util import unique
from Orange.widgets import gui
from Orange.widgets.gui import HorizontalGridDelegate
from Orange.widgets.settings import ContextSetting
from Orange.widgets.utils.itemmodels import TableModel


class Column:
    name = 0
    tpe = 1
    place = 2
    values = 3
    not_valid = 4


class Place:
    feature = 0
    class_var = 1
    meta = 2
    skip = 3


class VarTableModel(QAbstractTableModel):
    DISCRETE_VALUE_DISPLAY_LIMIT = 20

    places = "feature", "target", "meta", "skip"
    typenames = "categorical", "numeric", "text", "datetime"
    vartypes = DiscreteVariable, ContinuousVariable, StringVariable, TimeVariable
    name2type = dict(zip(typenames, vartypes))
    type2name = dict(zip(vartypes, typenames))

    def __init__(self, variables, *args):
        super().__init__(*args)
        self.variables = variables
        self.orig_variables = None
        self.set_orig_variables(variables)

    def set_orig_variables(self, variables):
        self.orig_variables = deepcopy(variables)

    def reset_variables(self):
        self.set_variables(deepcopy(self.orig_variables))

    def set_variables(self, variables):
        self.modelAboutToBeReset.emit()
        self.variables = variables
        self.modelReset.emit()

    def rowCount(self, parent):
        return 0 if parent.isValid() else len(self.variables)

    @staticmethod
    def columnCount(parent):
        return 0 if parent.isValid() else Column.not_valid

    def data(self, index, role):
        row, col = index.row(), index.column()
        val = self.variables[row][col]
        if role in (Qt.DisplayRole, Qt.EditRole):
            if col == Column.tpe:
                return self.type2name[val]
            if col == Column.place:
                return self.places[val]
            else:
                return val
        if role == Qt.DecorationRole:
            if col == Column.tpe:
                return gui.attributeIconDict[self.vartypes.index(val) + 1]
        if role == Qt.ForegroundRole:
            if self.variables[row][Column.place] == Place.skip \
                    and col != Column.place:
                return QColor(160, 160, 160)
        if role == Qt.BackgroundRole:
            place = self.variables[row][Column.place]
            mapping = [Place.meta, Place.feature, Place.class_var, None]
            return TableModel.ColorForRole.get(mapping[place], None)
        if role == Qt.FontRole:
            if self.variables[row] != self.orig_variables[row]:
                font = QFont()
                font.setBold(True)
                return font
        return None

    def setData(self, index, value, role=Qt.EditRole):
        row, col = index.row(), index.column()
        row_data = self.variables[row]
        if role == Qt.EditRole:
            if col == Column.name and not (value.isspace() or value == ""):
                row_data[col] = value
            elif col == Column.tpe:
                vartype = self.name2type[value]
                row_data[col] = vartype
                if not vartype.is_primitive() and \
                                row_data[Column.place] < Place.meta:
                    row_data[Column.place] = Place.meta
            elif col == Column.place:
                row_data[col] = self.places.index(value)
            else:
                return False
            # Settings may change background colors
            self.dataChanged.emit(index.sibling(row, 0), index.sibling(row, 3))
            return True
        return False

    def headerData(self, i, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole and i < 4:
            return ("Name", "Type", "Role", "Values")[i]
        if role == Qt.TextAlignmentRole:
            return Qt.AlignLeft
        return super().headerData(i, orientation, role)

    def flags(self, index):
        if index.column() == Column.values:
            return super().flags(index)
        return super().flags(index) | Qt.ItemIsEditable


class ComboDelegate(HorizontalGridDelegate):
    def __init__(self, view, items):
        super().__init__()
        self.view = view
        self.items = items

    def createEditor(self, parent, _option, index):
        # This ugly hack closes the combo when the user selects an item
        class Combo(QComboBox):
            def __init__(self, *args):
                super().__init__(*args)
                self.popup_shown = False
                self.highlighted_text = None

            def highlight(self, index):
                self.highlighted_text = index

            def showPopup(self, *args):
                super().showPopup(*args)
                self.popup_shown = True

            # Here, we need `self` from the closure
            # pylint: disable=no-self-argument,attribute-defined-outside-init
            def hidePopup(me):
                if me.popup_shown:
                    self.view.model().setData(
                        index, me.highlighted_text, Qt.EditRole)
                    self.popup_shown = False
                super().hidePopup()
                self.view.closeEditor(me, self.NoHint)

        combo = Combo(parent)
        combo.highlighted[str].connect(combo.highlight)
        return combo


class VarTypeDelegate(ComboDelegate):
    def setEditorData(self, combo, index):
        combo.clear()
        no_numeric = not self.view.model().variables[
            index.row()][Column.not_valid]
        if no_numeric:
            # Do not allow selection of numeric and datetime
            items = [i for i in self.items if i not in ("numeric", "datetime")]
        else:
            items = self.items

        ind = items.index(index.data())
        combo.addItems(items)
        combo.setCurrentIndex(ind)
        combo.showPopup()


class PlaceDelegate(ComboDelegate):
    def setEditorData(self, combo, index):
        combo.clear()
        to_meta = not self.view.model().variables[
            index.row()][Column.tpe].is_primitive()
        combo.addItems(self.items[2 * to_meta:])
        combo.setCurrentIndex(self.items.index(index.data()) - 2 * to_meta)
        combo.showPopup()


class DomainEditor(QTableView):
    """Component for editing of variable types.

    Parameters
    ----------
    widget : parent widget
    """

    variables = ContextSetting([])

    def __init__(self, widget):
        super().__init__()
        widget.settingsHandler.initialize(self)
        widget.contextAboutToBeOpened.connect(lambda args: self.set_domain(args[0]))
        widget.contextOpened.connect(lambda: self.model().set_variables(self.variables))
        widget.contextClosed.connect(lambda: self.model().set_variables([]))

        self.setModel(VarTableModel(self.variables, self))

        self.setSelectionMode(QTableView.NoSelection)
        self.horizontalHeader().setStretchLastSection(True)
        self.setShowGrid(False)
        self.setEditTriggers(
            QTableView.SelectedClicked | QTableView.DoubleClicked)
        self.setSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        # setItemDelegate(ForColumn) apparently does not take ownership
        # of delegates, and coredumps if a a references is not stored somewhere.
        # We thus store delegates as attributes
        self.grid_delegate = HorizontalGridDelegate()
        self.setItemDelegate(self.grid_delegate)
        self.vartype_delegate = VarTypeDelegate(self, VarTableModel.typenames)
        self.setItemDelegateForColumn(Column.tpe, self.vartype_delegate)
        self.place_delegate = PlaceDelegate(self, VarTableModel.places)
        self.setItemDelegateForColumn(Column.place, self.place_delegate)

    @staticmethod
    def _is_missing(x):
        return str(x) in ("nan", "")

    @staticmethod
    def _iter_vals(x):
        """Iterate over values of sparse or dense arrays."""
        for i in range(x.shape[0]):
            yield x[i, 0]

    @staticmethod
    def _to_column(x, to_sparse, dtype=None):
        """Transform list of values to sparse/dense column array."""
        x = np.asarray(x, dtype=dtype).reshape(-1, 1)
        if to_sparse:
            x = sp.csc_matrix(x)
        return x

    @staticmethod
    def _merge(cols, force_dense=False):
        if len(cols) == 0:
            return None

        all_dense = not any(sp.issparse(c) for c in cols)
        if all_dense:
            return np.hstack(cols)
        if force_dense:
            return np.hstack([c.toarray() if sp.issparse(c) else c for c in cols])
        sparse_cols = [c if sp.issparse(c) else sp.csc_matrix(c) for c in cols]
        return sp.hstack(sparse_cols).tocsr()

    def get_domain(self, domain, data, deduplicate=False):
        """
        Create domain (and dataset) from changes made in the widget.

        Returns
        -------

        Args:
            domain (Domain): original domain
            data (Table): original data
            deduplicate (bool): if True, variable names are deduplicated and
               the result contains an additional list with names of renamed
               variables

        Returns:
            (new_domain, [attribute_columns, class_var_columns, meta_columns])
            or
            (new_domain, [attribute_columns, class_var_columns, meta_columns], renamed)
        """
        # Allow type-checking with type() instead of isinstance() for exact comparison
        # pylint: disable=unidiomatic-typecheck
        variables = self.model().variables
        places = [[], [], []]  # attributes, class_vars, metas
        cols = [[], [], []]  # Xcols, Ycols, Mcols

        def numbers_are_round(var, col_data):
            if type(var) == ContinuousVariable:
                data = np.asarray(col_data.data)  # Works for dense and sparse
                data = data[~np.isnan(data)]
                return (data == data.astype(int)).all()
            return False

        # Exit early with original domain if the user didn't actually change anything
        if all((name == orig_var.name and tpe == type(orig_var) and place == orig_plc)
               for (name, tpe, place, _, _), (orig_var, orig_plc) in
               zip(variables,
                   chain(((at, Place.feature) for at in domain.attributes),
                         ((cl, Place.class_var) for cl in domain.class_vars),
                         ((mt, Place.meta) for mt in domain.metas)))):
            if deduplicate:
                return domain, [data.X, data.Y, data.metas], []
            else:
                return domain, [data.X, data.Y, data.metas]

        relevant_names = [var[0] for var in variables if var[2] != Place.skip]
        if deduplicate:
            renamed_iter = iter(get_unique_names_duplicates(relevant_names))
        else:
            renamed_iter = iter(relevant_names)
        renamed = []
        for (name, tpe, place, _, may_be_numeric), (orig_var, orig_plc) in \
                zip(variables,
                        chain([(at, Place.feature) for at in domain.attributes],
                              [(cl, Place.class_var) for cl in domain.class_vars],
                              [(mt, Place.meta) for mt in domain.metas])):
            if place == Place.skip:
                continue

            new_name = next(renamed_iter)
            if new_name != name and name not in renamed:
                renamed.append(name)

            col_data = self._get_column(data, orig_var, orig_plc)
            is_sparse = sp.issparse(col_data)

            if new_name == orig_var.name and tpe == type(orig_var):
                var = orig_var
            elif tpe == type(orig_var):
                var = orig_var.copy(name=new_name)
            elif tpe == DiscreteVariable:
                values = natural_sorted(
                    list(str(i) for i in unique(col_data)
                         if not self._is_missing(i))
                )
                round_numbers = numbers_are_round(orig_var, col_data)
                col_data = [np.nan if self._is_missing(x) else values.index(str(x))
                            for x in self._iter_vals(col_data)]
                if round_numbers:
                    values = [str(int(float(v))) for v in values]
                var = tpe(new_name, values)
                col_data = self._to_column(col_data, is_sparse)
            elif tpe == StringVariable:
                var = tpe.make(new_name)
                if type(orig_var) in [DiscreteVariable, TimeVariable]:
                    col_data = [orig_var.repr_val(x) if not np.isnan(x) else ""
                                for x in self._iter_vals(col_data)]
                elif type(orig_var) == ContinuousVariable:
                    round_numbers = numbers_are_round(orig_var, col_data)
                    col_data = ['' if np.isnan(x) else
                                str(int(x)) if round_numbers else
                                orig_var.repr_val(x)
                                for x in self._iter_vals(col_data)]
                # don't obey sparsity for StringVariable since they are
                # in metas which are transformed to dense below
                col_data = self._to_column(col_data, False, dtype=object)
            elif tpe == ContinuousVariable and type(orig_var) == DiscreteVariable:
                var = tpe.make(new_name)
                if may_be_numeric:
                    col_data = [np.nan if self._is_missing(x) else float(orig_var.values[int(x)])
                                for x in self._iter_vals(col_data)]
                col_data = self._to_column(col_data, is_sparse)
            else:
                var = tpe(new_name)
            places[place].append(var)
            cols[place].append(col_data)

        # merge columns for X, Y and metas
        feats = cols[Place.feature]
        X = self._merge(feats) if feats else np.empty((len(data), 0))
        Y = self._merge(cols[Place.class_var], force_dense=True)
        m = self._merge(cols[Place.meta], force_dense=True)
        domain = Domain(*places)
        if deduplicate:
            return domain, [X, Y, m], renamed
        else:
            return domain, [X, Y, m]

    @staticmethod
    def _get_column(data, source_var, source_place):
        """ Extract column from data and preserve sparsity. """
        if source_place == Place.meta:
            col_data = data[:, source_var].metas
        elif source_place == Place.class_var:
            col_data = data[:, source_var].Y.reshape(-1, 1)
        else:
            col_data = data[:, source_var].X
        return col_data

    def set_domain(self, domain):
        self.variables = self.parse_domain(domain)
        self.model().set_orig_variables(self.variables)

    def reset_domain(self):
        self.model().reset_variables()
        self.variables = self.model().variables

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            self.edit(index)
            event.accept()
        else:
            super().mousePressEvent(event)

    @staticmethod
    def parse_domain(domain):
        """Convert domain into variable representation used by
        the VarTableModel.

        Parameters
        ----------
        domain : the domain to convert

        Returns
        -------
        list of [variable_name, var_type, place, values, can_be_numeric] lists.

        """
        if domain is None:
            return []

        def may_be_numeric(var):
            if var.is_continuous:
                return True
            if var.is_discrete:
                try:
                    sum(float(x) for x in var.values)
                    return True
                except ValueError:
                    return False
            return False

        def discrete_value_display(value_list):
            result = ", ".join(str(v)
                               for v in value_list[:VarTableModel.DISCRETE_VALUE_DISPLAY_LIMIT])
            if len(value_list) > VarTableModel.DISCRETE_VALUE_DISPLAY_LIMIT:
                result += ", ..."
            return result

        return [
            [var.name, type(var), place,
             discrete_value_display(var.values) if var.is_discrete else "",
             may_be_numeric(var)]
            for place, vars in enumerate(
                (domain.attributes, domain.class_vars, domain.metas))
            for var in vars
        ]



import inspect


class WrapperMeta(type):
    """
    Meta class for scikit-learn wrapper classes.

    This is used for docstring generation/templating upon
    class definition.
    The client (class using this meta class) should define a class
    attribute `__wraps__` which contains the wrapped class.

    For instance::

        >>> class Bar:
        ...    '''I am a Bar'''

        >>> class Foo(metaclass=WrapperMeta):
        ...     __wraps__ = Bar

        >>> Foo.__doc__
        '...A wrapper for ...Bar`. The following is its
            documentation:...I am a Bar...'

    The client can also define a template for the docstring

        >>> class Foo(metaclass=WrapperMeta):
        ...     '''
        ...     Here is what ${sklname} says about itself:
        ...     ${skldoc}
        ...     '''
        ...     __wraps__ = Bar

        >>> Bar.__doc__
        'I am a Bar'

        >>> Foo.__doc__
        '...Here is what ...Bar says about itself:...I am a Bar...'

    """
    def __new__(cls, name, bases, dict_):
        cls = type.__new__(cls, name, bases, dict_)
        wrapped = getattr(cls, "__wraps__", getattr(cls, "__wrapped__", None))
        if wrapped is not None:
            doc = cls.__doc__ or """
A wrapper for `${sklname}`. The following is its documentation:

${skldoc}
            """
            sklname = "{}.{}".format(inspect.getmodule(wrapped).__name__,
                                     wrapped.__name__)
            skldoc = inspect.getdoc(wrapped) or ''
            # FIXME: make sure skl-extended classes are API-compatible
            if "Attributes\n---------" in skldoc:
                skldoc = skldoc[:skldoc.index('Attributes\n---------')]
            if "Examples\n--------" in skldoc:
                skldoc = skldoc[:skldoc.index('Examples\n--------')]
            if "Parameters\n---------" in skldoc:
                skldoc = skldoc[:skldoc.index('Parameters\n---------')]
            cls.__doc__ = (doc
                           .replace('${sklname}', sklname)
                           .replace('${skldoc}', inspect.cleandoc(skldoc)))
        return cls



from Orange.data import ContinuousVariable
from .base import Benchmark, benchmark


class BenchContinuous(Benchmark):

    # pylint: disable=no-self-use
    @benchmark()
    def bench_str_val_decimals(self):
        a = ContinuousVariable("a", 4)
        for _ in range(1000):
            a.str_val(1.23456)

    # pylint: disable=no-self-use
    @benchmark()
    def bench_str_val_g(self):
        a = ContinuousVariable("a")
        for _ in range(1000):
            a.str_val(1.23456)



# pylint: disable=protected-access
import unittest
from unittest.mock import Mock

import numpy as np

from Orange.data import ContinuousVariable, TimeVariable, \
    DiscreteVariable, StringVariable, Table
from Orange.data.io_base import _TableHeader, _TableBuilder, DataTableMixin


class InitTestData(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.header0 = [["0.1", "0.5", "0.1", "21.0"],
                       ["0.2", "0.1", "2.5", "123.0"],
                       ["0.0", "0.0", "0.0", "0.0"]]
        cls.header1 = [["a", "b", "c", "d"],
                       ["red", "2019-10-10", "0.0", "21.0"],
                       ["red", "2019-10-12", "0.0", "123.0"],
                       ["green", "2019-10-11", "", "0.0"]]
        cls.header1_flags = [["m#a", "cC#b", "m#c", "d", "i#e", "f"],
                             ["red", "0.5", "0.0", "0.0", "aa", "a"],
                             ["red", "0.1", "1.0", "1.0", "b", "b"],
                             ["green", "0.0", "2.0", "2.0", "c", "c"]]
        cls.header3 = [["a", "b", "c", "d", "w", "e", "f", "g"],
                       ["d", "c", "c", "c", "c", "d", "s", "yes no"],
                       ["meta", "class", "meta", "", "weight", "i", "", ""],
                       ["red", "0.5", "0.0", "0.0", "0.5", "aa", "a", "no"]]


class TestTableHeader(InitTestData):
    def test_rename_variables(self):
        th = _TableHeader([["a", "", "b", "", "a"]])
        self.assertListEqual(th.names, ["a (1)", "", "b", "", "a (2)"])

    def test_get_header_data_0(self):
        names, types, flags = _TableHeader.create_header_data([])
        self.assertListEqual(names, [])
        self.assertListEqual(types, [])
        self.assertListEqual(flags, [])

    def test_get_header_data_1(self):
        names, types, flags = _TableHeader.create_header_data(self.header1[:1])
        self.assertListEqual(names, ["a", "b", "c", "d"])
        self.assertListEqual(types, ["", "", "", ""])
        self.assertListEqual(flags, ["", "", "", ""])

    def test_get_header_data_1_flags(self):
        names, types, flags = _TableHeader.create_header_data(
            self.header1_flags[:1])
        self.assertListEqual(names, ["a", "b", "c", "d", "e", "f"])
        self.assertListEqual(types, ["", "c", "", "", "", ""])
        self.assertListEqual(flags, ["m", "c", "m", "", "i", ""])

    def test_get_header_data_3(self):
        names, types, flags = _TableHeader.create_header_data(self.header3[:3])
        self.assertListEqual(names, ["a", "b", "c", "d", "w", "e", "f", "g"])
        self.assertListEqual(
            types, ["d", "c", "c", "c", "c", "d", "s", "yes no"])
        self.assertListEqual(
            flags, ["meta", "class", "meta", "", "weight", "i", "", ""])


class TestTableBuilder(InitTestData):
    def test_string_column(self):
        data = np.array(self.header1[1:])
        creator = _TableBuilder._get_column_creator("s")
        column = creator(data, 0)

        self.assertIsNone(column.valuemap)
        np.testing.assert_array_equal(column.values, ["red", "red", "green"])
        np.testing.assert_array_equal(column.orig_values,
                                      ["red", "red", "green"])
        self.assertEqual(column.coltype, StringVariable)
        self.assertDictEqual(column.coltype_kwargs, {})

    def test_continuous_column(self):
        data = np.array(self.header0)
        creator = _TableBuilder._get_column_creator("c")
        column = creator(data, 0)

        self.assertIsNone(column.valuemap)
        np.testing.assert_array_equal(column.values, [0.1, 0.2, 0])
        np.testing.assert_array_equal(column.orig_values,
                                      ["0.1", "0.2", "0.0"])
        self.assertEqual(column.coltype, ContinuousVariable)
        self.assertDictEqual(column.coltype_kwargs, {})

    def test_continuous_column_raises(self):
        data = np.array([["a", "2"], ["3", "4"]])
        creator = _TableBuilder._get_column_creator("c")
        self.assertRaises(ValueError, creator, data, 0)

    def test_time_column(self):
        data = np.array(self.header1[1:])
        creator = _TableBuilder._get_column_creator("t")
        column = creator(data, 1)

        self.assertIsNone(column.valuemap)
        np.testing.assert_array_equal(
            column.values, ['2019-10-10', '2019-10-12', '2019-10-11'])
        np.testing.assert_array_equal(
            column.orig_values, ['2019-10-10', '2019-10-12', '2019-10-11'])
        self.assertEqual(column.coltype, TimeVariable)
        self.assertDictEqual(column.coltype_kwargs, {})

    def test_discrete_column(self):
        data = np.array(self.header1[1:])
        creator = _TableBuilder._get_column_creator("d")
        columns = creator(data, 0)

        self.assertListEqual(columns.valuemap, ['green', 'red'])
        np.testing.assert_array_equal(columns.values, ["red", "red", "green"])
        np.testing.assert_array_equal(columns.orig_values,
                                      ["red", "red", "green"])
        self.assertEqual(columns.coltype, DiscreteVariable)
        self.assertDictEqual(columns.coltype_kwargs, {})

    def test_column_parts_discrete_values(self):
        data = np.array(self.header1[1:])
        vals = "green red"
        creator = _TableBuilder._get_column_creator(vals)
        column = creator(data, 0, vals)

        self.assertListEqual(column.valuemap, ['green', 'red'])
        np.testing.assert_array_equal(column.values, ["red", "red", "green"])
        np.testing.assert_array_equal(column.orig_values,
                                      ["red", "red", "green"])
        self.assertEqual(column.coltype, DiscreteVariable)
        self.assertDictEqual(column.coltype_kwargs, {})

    def test_unknown_type_column(self):
        data = np.array(self.header0)
        creator = _TableBuilder._get_column_creator("")
        column = creator(data, 0)

        self.assertIsNone(column.valuemap)
        np.testing.assert_array_equal(column.values, [0.1, 0.2, 0])
        np.testing.assert_array_equal(column.orig_values, ["0.1", "0.2", "0.0"])
        self.assertEqual(column.coltype, ContinuousVariable)
        self.assertDictEqual(column.coltype_kwargs, {})


class TestDataTableMixin(InitTestData):
    def test_data_table_empty(self):
        self.assertIsInstance(DataTableMixin.data_table([]), Table)

    def test_data_table_0(self):
        self.assertIsInstance(DataTableMixin.data_table(self.header0), Table)

    def test_data_table_1(self):
        self.assertIsInstance(DataTableMixin.data_table(self.header1), Table)

    def test_data_table_1_flags(self):
        self.assertIsInstance(DataTableMixin.data_table(
            self.header1_flags), Table)

    def test_data_table_3(self):
        self.assertIsInstance(DataTableMixin.data_table(self.header3), Table)

    def test_parse_headers_empty(self):
        headers, data = DataTableMixin.parse_headers([])
        self.assertListEqual(headers, [])
        self.assertListEqual(list(data), [])

    def test_parse_headers_0(self):
        hdata = self.header0
        headers, data = DataTableMixin.parse_headers(hdata)
        self.assertListEqual(headers, [])
        self.assertListEqual(list(data), hdata)

    def test_parse_headers_1(self):
        hdata = self.header1
        headers, data = DataTableMixin.parse_headers(hdata)
        self.assertListEqual(headers, [["a", "b", "c", "d"]])
        self.assertListEqual(list(data), hdata[1:])

    def test_parse_headers_1_flags(self):
        hdata = self.header1_flags
        headers, data = DataTableMixin.parse_headers(hdata)
        self.assertListEqual(
            headers, [["m#a", "cC#b", "m#c", "d", "i#e", "f"]])
        self.assertListEqual(list(data), hdata[1:])

    def test_parse_headers_3(self):
        hdata = self.header3
        headers, data = DataTableMixin.parse_headers(hdata)
        self.assertListEqual(
            headers, [["a", "b", "c", "d", "w", "e", "f", "g"],
                      ["d", "c", "c", "c", "c", "d", "s", "yes no"],
                      ["meta", "class", "meta", "", "weight", "i", "", ""]])
        self.assertListEqual(list(data), hdata[3:])

    def test_adjust_data_width_lengthen(self):
        names, types, flags = ["a", "b", "c", "d", "e"], ["", "c"], ["m", "c"]
        header = Mock()
        header.names = names
        header.types = types
        header.flags = flags
        adjusted, n = DataTableMixin.adjust_data_width(self.header0, header)
        _data = np.hstack((np.array(self.header0, dtype=object),
                           np.array([[""]] * 3, dtype=object)))
        np.testing.assert_array_equal(_data, adjusted)
        self.assertEqual(adjusted.shape, (len(self.header0), 5))
        self.assertEqual(n, 5)
        self.assertListEqual(names, ["a", "b", "c", "d", "e"])
        self.assertListEqual(types, ["", "c", "", "", ""])
        self.assertListEqual(flags, ["m", "c", "", "", ""])

    def test_adjust_data_width_shorten(self):
        names, types, flags = ["a", "b", "c"], ["", "c"], ["m", "c"]
        header = Mock()
        header.names = names
        header.types = types
        header.flags = flags
        with self.assertWarns(UserWarning):
            adjusted, n = DataTableMixin.adjust_data_width(self.header0, header)
        np.testing.assert_array_equal(
            adjusted, np.array(self.header0, dtype=object)[:, :3])
        self.assertEqual(adjusted.shape, (len(self.header0), 3))
        self.assertEqual(n, 3)
        self.assertListEqual(names, ["a", "b", "c"])
        self.assertListEqual(types, ["", "c", ""])
        self.assertListEqual(flags, ["m", "c", ""])

    def test_adjust_data_width_empty(self):
        names, types, flags = ["a", "b"], [], []
        header = Mock()
        header.names = names
        header.types = types
        header.flags = flags
        data = [["", ""], ["", ""]]
        adjusted, n = DataTableMixin.adjust_data_width(data, header)
        np.testing.assert_array_equal(adjusted, [])
        self.assertEqual(n, 2)
        self.assertListEqual(names, ["a", "b"])
        self.assertListEqual(types, [])
        self.assertListEqual(flags, [])


if __name__ == '__main__':
    unittest.main()



import unittest


class TestOrange(unittest.TestCase):
    def test_orange_has_modules(self):
        import pkgutil
        import Orange
        unimported = ['canvas', 'datasets', 'testing', 'tests', 'setup',
                      'util', 'widgets']
        for _, name, __ in pkgutil.iter_modules(Orange.__path__):
            if name not in unimported:
                self.assertIn(name, Orange.__dict__)



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring,unsubscriptable-object
import unittest
from unittest.mock import Mock

from Orange.data import Table
from Orange.widgets.data.owdatasampler import OWDataSampler
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.utils.state_summary import format_summary_details


class TestOWDataSampler(WidgetTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = Table("iris")
        cls.zoo = Table("zoo")

    def setUp(self):
        self.widget = self.create_widget(OWDataSampler)  # type: OWDataSampler

    def test_error_message(self):
        """ Check if error message appears and then disappears when
        data is removed from input"""
        self.widget.controls.sampling_type.buttons[2].click()
        self.send_signal("Data", self.iris)
        self.assertFalse(self.widget.Error.too_many_folds.is_shown())
        self.send_signal("Data", self.iris[:5])
        self.assertTrue(self.widget.Error.too_many_folds.is_shown())
        self.send_signal("Data", None)
        self.assertFalse(self.widget.Error.too_many_folds.is_shown())
        self.send_signal("Data", Table.from_domain(self.iris.domain))
        self.assertTrue(self.widget.Error.no_data.is_shown())

    def test_stratified_on_unbalanced_data(self):
        unbalanced_data = self.iris[:51]

        self.widget.controls.stratify.setChecked(True)
        self.send_signal("Data", unbalanced_data)
        self.assertTrue(self.widget.Warning.could_not_stratify.is_shown())

    def test_bootstrap(self):
        self.select_sampling_type(self.widget.Bootstrap)

        self.send_signal("Data", self.iris)

        in_input = set(self.iris.ids)
        sample = self.get_output("Data Sample")
        in_sample = set(sample.ids)
        in_remaining = set(self.get_output("Remaining Data").ids)

        # Bootstrap should sample len(input) instances
        self.assertEqual(len(sample), len(self.iris))
        # Sample and remaining should cover all instances, while none
        # should be present in both
        self.assertEqual(len(in_sample | in_remaining), len(in_input))
        self.assertEqual(len(in_sample & in_remaining), 0)
        # Sampling with replacement will always produce at least one distinct
        # instance in sample, and at least one instance in remaining with
        # high probability (1-(1/150*2/150*...*150/150) ~= 1-2e-64)
        self.assertGreater(len(in_sample), 0)
        self.assertGreater(len(in_remaining), 0)

    def select_sampling_type(self, sampling_type):
        buttons = self.widget.controls.sampling_type.group.buttons()
        buttons[sampling_type].click()

    def test_no_intersection_in_outputs(self):
        """ Check whether outputs intersect and whether length of outputs sums
        to length of original data"""
        self.send_signal("Data", self.iris)
        w = self.widget
        sampling_types = [w.FixedProportion, w.FixedSize, w.CrossValidation]

        for replicable in [True, False]:
            for stratified in [True, False]:
                for sampling_type in sampling_types:
                    self.widget.cb_seed.setChecked(replicable)
                    self.widget.cb_stratify.setChecked(stratified)
                    self.select_sampling_type(sampling_type)
                    self.widget.commit()

                    sample = self.get_output("Data Sample")
                    other = self.get_output("Remaining Data")
                    self.assertEqual(len(self.iris), len(sample) + len(other))
                    self.assertNoIntersection(sample, other)

    def test_bigger_size_with_replacement(self):
        """Allow bigger output without replacement."""
        self.send_signal('Data', self.iris[:2])
        sample_size = self.set_fixed_sample_size(3, with_replacement=True)
        self.assertEqual(3, sample_size, 'Should be able to set a bigger size '
                         'with replacement')

    def test_bigger_size_without_replacement(self):
        """Lower output samples to match input's without replacement."""
        self.send_signal('Data', self.iris[:2])
        sample_size = self.set_fixed_sample_size(3)
        self.assertEqual(2, sample_size)

    def test_bigger_output_warning(self):
        """Should warn when sample size is bigger than input."""
        self.send_signal('Data', self.iris[:2])
        self.set_fixed_sample_size(3, with_replacement=True)
        self.assertTrue(self.widget.Warning.bigger_sample.is_shown())

    def test_shuffling(self):
        self.send_signal('Data', self.iris)

        self.set_fixed_sample_size(150)
        self.assertFalse(self.widget.Warning.bigger_sample.is_shown())
        sample = self.get_output("Data Sample")
        self.assertTrue((self.iris.ids != sample.ids).any())
        self.assertEqual(set(self.iris.ids), set(sample.ids))

        self.select_sampling_type(self.widget.FixedProportion)
        self.widget.sampleSizePercentage = 100
        self.widget.commit()
        sample = self.get_output("Data Sample")
        self.assertTrue((self.iris.ids != sample.ids).any())
        self.assertEqual(set(self.iris.ids), set(sample.ids))

    def test_summary(self):
        """Check if status bar is updated when data is received"""
        data = self.iris
        input_sum = self.widget.info.set_input_summary = Mock()
        output_sum = self.widget.info.set_output_summary = Mock()

        self.send_signal(self.widget.Inputs.data, data)
        input_sum.assert_called_with(len(data), format_summary_details(data))
        output = self.get_output(self.widget.Outputs.data_sample)
        output_sum.assert_called_with(len(output),
                                      format_summary_details(output))

        input_sum.reset_mock()
        output_sum.reset_mock()
        self.send_signal(self.widget.Inputs.data, None)
        input_sum.assert_called_once()
        self.assertEqual(input_sum.call_args[0][0].brief, "")
        output_sum.assert_called_once()
        self.assertEqual(output_sum.call_args[0][0].brief, "")

    def set_fixed_sample_size(self, sample_size, with_replacement=False):
        """Set fixed sample size and return the number of gui spin.

        Return the actual number in gui so we can check whether it is different
        from sample_size. The number can be changed depending on the spin
        maximum value.
        """
        self.select_sampling_type(self.widget.FixedSize)
        self.widget.controls.replacement.setChecked(with_replacement)
        self.widget.sampleSizeSpin.setValue(sample_size)
        self.widget.commit()
        return self.widget.sampleSizeSpin.value()

    def assertNoIntersection(self, sample, other):
        self.assertFalse(bool(set(sample.ids) & set(other.ids)))

    def test_cv_outputs(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.iris)

        self.select_sampling_type(w.CrossValidation)
        self.widget.commit()
        self.assertEqual(len(self.get_output(w.Outputs.data_sample)), 135)
        self.assertEqual(len(self.get_output(w.Outputs.remaining_data)), 15)

    def test_cv_output_migration(self):
        self.assertFalse(self.widget.compatibility_mode)

        settings = {"sampling_type": OWDataSampler.CrossValidation}
        OWDataSampler.migrate_settings(settings, version=2)
        self.assertFalse(settings.get("compatibility_mode", False))

        settings = {"sampling_type": OWDataSampler.FixedProportion}
        OWDataSampler.migrate_settings(settings, version=1)
        self.assertFalse(settings.get("compatibility_mode", False))

        settings = {"sampling_type": OWDataSampler.CrossValidation}
        OWDataSampler.migrate_settings(settings, version=1)
        self.assertTrue(settings["compatibility_mode"])

        w = self.create_widget(
            OWDataSampler,
            stored_settings={"sampling_type": OWDataSampler.CrossValidation,
                             "__version__": 1})
        self.assertTrue(w.compatibility_mode)

        self.send_signal(w.Inputs.data, self.iris)
        self.select_sampling_type(w.CrossValidation)
        w.commit()
        self.assertEqual(len(self.get_output(w.Outputs.data_sample)), 15)
        self.assertEqual(len(self.get_output(w.Outputs.remaining_data)), 135)

    def test_send_report(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.iris)

        w.stratify = True
        w.use_seed = True

        self.select_sampling_type(0)
        w.commit()
        w.send_report()

        self.select_sampling_type(1)
        w.sampleSizeNumber = 1
        w.commit()
        w.send_report()

        w.sampleSizeNumber = 10
        w.replacement = False
        w.commit()
        w.send_report()

        w.replacement = True
        w.commit()
        w.send_report()

        self.select_sampling_type(2)
        w.commit()
        w.send_report()

        self.select_sampling_type(3)
        w.commit()
        w.send_report()


if __name__ == "__main__":
    unittest.main()



import asyncio
import json
import logging
import random
import uuid
from json import JSONDecodeError
from os import getenv
from typing import Any, Callable, List, Optional

from AnyQt.QtCore import QSettings
from httpx import AsyncClient, NetworkError, ReadTimeout, Response

from Orange.misc.utils.embedder_utils import (EmbedderCache,
                                              EmbeddingCancelledException,
                                              EmbeddingConnectionError,
                                              get_proxies)

log = logging.getLogger(__name__)


class ServerEmbedderCommunicator:
    """
    This class needs to be inherited by the class which re-implements
    _encode_data_instance and defines self.content_type. For sending a table
    with data items use embedd_table function. This one is called with the
    complete Orange data Table. Then _encode_data_instance needs to extract
    data to be embedded from the RowInstance. For images, it takes the image
    path from the table, load image, and transform it into bytes.

    Attributes
    ----------
    model_name
        The name of the model. Name is used in url to define what server model
        gets data to embedd and as a caching keyword.
    max_parallel_requests
        Number of image that can be sent to the server at the same time.
    server_url
        The base url of the server (without any additional subdomains)
    embedder_type
        The type of embedder (e.g. image). It is used as a part of url (e.g.
        when embedder_type is image url is api.garaza.io/image)
    """

    MAX_REPEATS = 3

    count_connection_errors = 0
    count_read_errors = 0
    max_errors = 10

    def __init__(
            self,
            model_name: str,
            max_parallel_requests: int,
            server_url: str,
            embedder_type: str,
    ) -> None:
        self.server_url = getenv("ORANGE_EMBEDDING_API_URL", server_url)
        self._model = model_name
        self.embedder_type = embedder_type

        # attribute that offers support for cancelling the embedding
        # if ran in another thread
        self._cancelled = False

        self.machine_id = None
        try:
            self.machine_id = QSettings().value(
                "error-reporting/machine-id", "", type=str
            ) or str(uuid.getnode())
        except TypeError:
            self.machine_id = str(uuid.getnode())
        self.session_id = str(random.randint(1, 1e10))

        self._cache = EmbedderCache(model_name)

        # default embedding timeouts are too small we need to increase them
        self.timeout = 180
        self.num_parallel_requests = 0
        self.max_parallel = max_parallel_requests
        self.content_type = None  # need to be set in a class inheriting

    def embedd_data(
            self,
            data: List[Any],
            processed_callback: Callable[[bool], None] = None,
    ) -> List[Optional[List[float]]]:
        """
        This function repeats calling embedding function until all items
        are embedded. It prevents skipped items due to network issues.
        The process is repeated for each item maximally MAX_REPEATS times.

        Parameters
        ----------
        data
            List with data that needs to be embedded.
        processed_callback
            A function that is called after each item is embedded
            by either getting a successful response from the server,
            getting the result from cache or skipping the item.

        Returns
        -------
        List of float list (embeddings) for successfully embedded
        items and Nones for skipped items.

        Raises
        ------
        EmbeddingConnectionError
            Error which indicate that the embedding is not possible due to
            connection error.
        EmbeddingCancelledException:
            If cancelled attribute is set to True (default=False).
        """
        # if there is less items than 10 connection error should be raised
        # earlier
        self.max_errors = min(len(data) * self.MAX_REPEATS, 10)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            embeddings = asyncio.get_event_loop().run_until_complete(
                self.embedd_batch(data, processed_callback)
            )
        except Exception:
            loop.close()
            raise

        loop.close()
        return embeddings

    async def embedd_batch(
            self, data: List[Any], proc_callback: Callable[[bool], None] = None
    ) -> List[Optional[List[float]]]:
        """
        Function perform embedding of a batch of data items.

        Parameters
        ----------
        data
            A list of data that must be embedded.
        proc_callback
            A function that is called after each item is fully processed
            by either getting a successful response from the server,
            getting the result from cache or skipping the item.

        Returns
        -------
        List of float list (embeddings) for successfully embedded
        items and Nones for skipped items.

        Raises
        ------
        EmbeddingCancelledException:
            If cancelled attribute is set to True (default=False).
        """
        requests = []
        async with AsyncClient(
                timeout=self.timeout, base_url=self.server_url, proxies=get_proxies()
        ) as client:
            for p in data:
                if self._cancelled:
                    raise EmbeddingCancelledException()
                requests.append(self._send_to_server(p, client, proc_callback))

            embeddings = await asyncio.gather(*requests)
        self._cache.persist_cache()
        assert self.num_parallel_requests == 0

        return embeddings

    async def __wait_until_released(self) -> None:
        while self.num_parallel_requests >= self.max_parallel:
            await asyncio.sleep(0.1)

    def __check_cancelled(self):
        if self._cancelled:
            raise EmbeddingCancelledException()

    async def _encode_data_instance(
            self, data_instance: Any
    ) -> Optional[bytes]:
        """
        The reimplementation of this function must implement the procedure
        to encode the data item in a string format that will be sent to the
        server. For images it is the byte string with an image. The encoding
        must be always equal for same data instance.

        Parameters
        ----------
        data_instance
            The row of an Orange data table.

        Returns
        -------
        Bytes encoding the data instance.
        """
        raise NotImplementedError

    async def _send_to_server(
            self,
            data_instance: Any,
            client: AsyncClient,
            proc_callback: Callable[[bool], None] = None,
    ) -> Optional[List[float]]:
        """
        Function get an data instance. It extract data from it and send them to
        server and retrieve responses.

        Parameters
        ----------
        data_instance
            Single row of the input table.
        client
            HTTPX client that communicates with the server
        proc_callback
            A function that is called after each item is fully processed
            by either getting a successful response from the server,
            getting the result from cache or skipping the item.

        Returns
        -------
        Embedding. For items that are not successfully embedded returns None.
        """
        await self.__wait_until_released()
        self.__check_cancelled()

        self.num_parallel_requests += 1
        # load bytes
        data_bytes = await self._encode_data_instance(data_instance)
        if data_bytes is None:
            self.num_parallel_requests -= 1
            return None

        # if data in cache return it
        cache_key = self._cache.md5_hash(data_bytes)
        emb = self._cache.get_cached_result_or_none(cache_key)

        if emb is None:
            # in case that embedding not sucessfull resend it to the server
            # maximally for MAX_REPEATS time
            for i in range(1, self.MAX_REPEATS + 1):
                self.__check_cancelled()
                url = (
                    f"/{self.embedder_type}/{self._model}?"
                    f"machine={self.machine_id}"
                    f"&session={self.session_id}&retry={i}"
                )
                emb = await self._send_request(client, data_bytes, url)
                if emb is not None:
                    self._cache.add(cache_key, emb)
                    break  # repeat only when embedding None
        if proc_callback:
            proc_callback(emb is not None)

        self.num_parallel_requests -= 1
        return emb

    async def _send_request(
            self, client: AsyncClient, data: bytes, url: str
    ) -> Optional[List[float]]:
        """
        This function sends a single request to the server.

        Parameters
        ----------
        client
            HTTPX client that communicates with the server
        data
            Data packed in the sequence of bytes.
        url
            Rest of the url string.

        Returns
        -------
        embedding
            Embedding. For items that are not successfully embedded returns
            None.
        """
        headers = {
            "Content-Type": self.content_type,
            "Content-Length": str(len(data)),
        }
        try:
            response = await client.post(url, headers=headers, data=data)
        except ReadTimeout as ex:
            log.debug("Read timeout", exc_info=True)
            # it happens when server do not respond in 60 seconds, in
            # this case we return None and items will be resend later

            # if it happens more than in ten consecutive cases it means
            # sth is wrong with embedder we stop embedding
            self.count_read_errors += 1

            if self.count_read_errors >= self.max_errors:
                self.num_parallel_requests = 0  # for safety reasons
                raise EmbeddingConnectionError from ex
            return None
        except (OSError, NetworkError) as ex:
            log.debug("Network error", exc_info=True)
            # it happens when no connection and items cannot be sent to the
            # server
            # we count number of consecutive errors
            # if more than 10 consecutive errors it means there is no
            # connection so we stop embedding with EmbeddingConnectionError
            self.count_connection_errors += 1
            if self.count_connection_errors >= self.max_errors:
                self.num_parallel_requests = 0  # for safety reasons
                raise EmbeddingConnectionError from ex
            return None
        except Exception:
            log.debug("Embedding error", exc_info=True)
            raise
        # we reset the counter at successful embedding
        self.count_connection_errors = 0
        self.count_read_errors = 0
        return self._parse_response(response)

    @staticmethod
    def _parse_response(response: Response) -> Optional[List[float]]:
        """
        This function get response and extract embeddings out of them.

        Parameters
        ----------
        response
            Response by the server

        Returns
        -------
        Embedding. For items that are not successfully embedded returns None.
        """
        if response.content:
            try:
                cont = json.loads(response.content.decode("utf-8"))
                return cont.get("embedding", None)
            except JSONDecodeError:
                # in case that embedding was not successful response is not
                # valid JSON
                return None
        else:
            return None

    def clear_cache(self):
        self._cache.clear_cache()

    def set_cancelled(self):
        self._cancelled = True
